from darq.registry import Registry
from darq.worker import Task


async def foobar():
    pass


async def foobar_2():
    pass


def test_registry():
    registry = Registry()
    task_name = 'foobar_task'
    worker_task = Task(
        name=task_name, coroutine=foobar, timeout_s=None, keep_result_s=None,
        max_tries=None, with_ctx=False,
    )
    task_name_2 = 'foobar_2_task'
    worker_task_2 = Task(
        name=task_name_2, coroutine=foobar_2, timeout_s=None,
        keep_result_s=None, max_tries=None, with_ctx=False,
    )

    registry.add(worker_task)
    registry.add(worker_task_2)
    assert registry.get(task_name) == worker_task
    assert registry.get(task_name_2) == worker_task_2
    assert set(registry.get_functions()) == {worker_task, worker_task_2}

    task_names = registry.get_function_names()
    assert set(task_names) == {task_name, task_name_2}



import asyncio
import functools
import logging
import typing as t
from dataclasses import dataclass
from datetime import datetime, timedelta
from operator import attrgetter
from ssl import SSLContext
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from uuid import uuid4

import aioredis
from aioredis import MultiExecError, Redis

from .constants import default_queue_name
from .constants import job_key_prefix
from .constants import result_key_prefix
from .jobs import deserialize_job
from .jobs import Deserializer
from .jobs import Job
from .jobs import JobDef
from .jobs import JobResult
from .jobs import serialize_job
from .jobs import Serializer
from .utils import timestamp_ms
from .utils import to_ms
from .utils import to_unix_ms

log = logging.getLogger('darq.connections')


@dataclass
class RedisSettings:
    """
    No-Op class used to hold redis connection redis_settings.

    Used by :func:`darq.connections.create_pool`
    and :class:`darq.worker.Worker`.
    """

    host: Union[str, List[Tuple[str, int]]] = 'localhost'
    port: int = 6379
    database: int = 0
    password: Optional[str] = None
    ssl: Union[bool, None, SSLContext] = None
    conn_timeout: int = 1
    conn_retries: int = 5
    conn_retry_delay: int = 1

    sentinel: bool = False
    sentinel_master: str = 'mymaster'
    sentinel_timeout: float = 0.2

    def __repr__(self) -> str:
        return '<RedisSettings {}>'.format(
            ' '.join(f'{k}={v}' for k, v in self.__dict__.items()),
        )


# extra time after the job is expected to start when the job key should expire,
# 1 day in ms
expires_extra_ms = 86_400_000


class ArqRedis(Redis):
    """
    Thin subclass of ``aioredis.Redis`` which adds
    :func:`darq.connections.enqueue_job`.

    :param redis_settings: an instance of ``darq.connections.RedisSettings``.
    :param job_serializer: a function that serializes Python objects to bytes,
                           defaults to pickle.dumps
    :param job_deserializer: a function that deserializes bytes into Python
                             objects, defaults to pickle.loads
    :param kwargs: keyword arguments directly passed to ``aioredis.Redis``.
    """

    def __init__(
        self,
        pool_or_conn: Union[aioredis.ConnectionsPool, aioredis.RedisConnection],
        job_serializer: Optional[Serializer] = None,
        job_deserializer: Optional[Deserializer] = None,
        **kwargs: Dict[str, t.Any],
    ) -> None:
        self.job_serializer = job_serializer
        self.job_deserializer = job_deserializer
        super().__init__(pool_or_conn, **kwargs)

    async def enqueue_job(
        self,
        function: str,
        args: Sequence[Any],
        kwargs: Mapping[str, Any],
        *,
        job_id: Optional[str] = None,
        queue_name: Optional[str] = None,
        defer_until: Optional[datetime] = None,
        defer_by: Union[None, int, float, timedelta] = None,
        expires: Union[None, int, float, timedelta] = None,
        job_try: Optional[int] = None,
    ) -> Optional[Job]:
        """
        Enqueue a job.

        :param function: Name of the function to call
        :param args: args to pass to the function
        :param kwargs: kwargs to pass to the function
        :param job_id: ID of the job, can be used to enforce job uniqueness
        :param queue_name: queue of the job, can be used to create job in
                           different queue
        :param defer_until: datetime at which to run the job
        :param defer_by: duration to wait before running the job
        :param expires: if the job still hasn't started after this duration,
                        do not run it
        :param job_try: useful when re-enqueueing jobs within a job
        :return: :class:`darq.jobs.Job` instance or ``None`` if a job with this
                 ID already exists
        """
        job_id = job_id or uuid4().hex
        job_key = job_key_prefix + job_id
        queue_name = queue_name or default_queue_name

        assert not (defer_until and defer_by), \
            "use either 'defer_until' or 'defer_by' or neither, not both"

        defer_by_ms = to_ms(defer_by)
        expires_ms = to_ms(expires)

        with await self as conn:
            pipe = conn.pipeline()
            pipe.unwatch()
            pipe.watch(job_key)
            job_exists = pipe.exists(job_key)
            job_result_exists = pipe.exists(result_key_prefix + job_id)
            await pipe.execute()
            if await job_exists or await job_result_exists:
                return None

            enqueue_time_ms = timestamp_ms()
            if defer_until is not None:
                score = to_unix_ms(defer_until)
            elif defer_by_ms:
                score = enqueue_time_ms + defer_by_ms
            else:
                score = enqueue_time_ms

            expires_ms = (
                expires_ms or score - enqueue_time_ms + expires_extra_ms
            )

            job = serialize_job(
                function, args, kwargs, job_try, enqueue_time_ms,
                serializer=self.job_serializer,
            )
            tr = conn.multi_exec()
            tr.psetex(job_key, expires_ms, job)
            tr.zadd(queue_name, score, job_id)
            try:
                await tr.execute()
            except MultiExecError:  # pragma: no cover
                # job got enqueued since we checked 'job_exists'
                # https://github.com/samuelcolvin/arq/issues/131
                # avoid warnings in log
                await asyncio.gather(*tr._results, return_exceptions=True)
                return None
        return Job(
            job_id, redis=self, _queue_name=queue_name,
            _deserializer=self.job_deserializer,
        )

    async def _get_job_result(self, key: str) -> JobResult:
        job_id = key[len(result_key_prefix):]
        job = Job(job_id, self, _deserializer=self.job_deserializer)
        r = t.cast(JobResult, await job.result_info())
        r.job_id = job_id
        return r

    async def all_job_results(self) -> List[JobResult]:
        """
        Get results for all jobs in redis.
        """
        keys = await self.keys(result_key_prefix + '*')
        results = await asyncio.gather(*[self._get_job_result(k) for k in keys])
        return sorted(results, key=attrgetter('enqueue_time'))

    async def _get_job_def(self, job_id: str, score: int) -> JobDef:
        v = await self.get(job_key_prefix + job_id, encoding=None)
        jd = deserialize_job(v, deserializer=self.job_deserializer)
        jd.score = score
        return jd

    async def queued_jobs(
            self, *, queue_name: str = default_queue_name,
    ) -> List[JobDef]:
        """
        Get information about queued, mostly useful when testing.
        """
        jobs = await self.zrange(queue_name, withscores=True)
        return await asyncio.gather(*[
            self._get_job_def(job_id, score) for job_id, score in jobs
        ])


async def create_pool(
    settings: RedisSettings = None,
    *,
    retry: int = 0,
    job_serializer: Optional[Serializer] = None,
    job_deserializer: Optional[Deserializer] = None,
) -> ArqRedis:
    """
    Create a new redis pool, retrying up to ``conn_retries`` times if the
    connection fails.

    Similar to ``aioredis.create_redis_pool`` except it returns a
    :class:`darq.connections.ArqRedis` instance, thus allowing job enqueuing.
    """
    settings = settings or RedisSettings()
    addr: t.Union[
        Tuple[str, int],
        List[Tuple[str, int]],
    ]

    if type(settings.host) is str and settings.sentinel:
        raise ValueError(
            "str provided for 'host' but 'sentinel' is true; "
            'list of sentinels expected',
        )

    if settings.sentinel:
        addr = t.cast(List[Tuple[str, int]], settings.host)

        async def pool_factory(
                *args: Any, **kwargs: Any,
        ) -> aioredis.sentinel.pool.SentinelPool:
            client = await aioredis.sentinel.create_sentinel_pool(
                *args,
                timeout=t.cast(RedisSettings, settings).sentinel_timeout,
                **kwargs,
            )
            return client.master_for(
                t.cast(RedisSettings, settings).sentinel_master,
            )

    else:
        pool_factory = functools.partial(
            aioredis.create_pool,
            create_connection_timeout=settings.conn_timeout,
        )
        addr = t.cast(str, settings.host), settings.port

    try:
        pool = await pool_factory(
            addr, db=settings.database, password=settings.password,
            ssl=settings.ssl, encoding='utf8',
        )
        pool = ArqRedis(
            pool,
            job_serializer=job_serializer, job_deserializer=job_deserializer,
        )
    except (OSError, aioredis.RedisError, asyncio.TimeoutError) as e:
        if retry < settings.conn_retries:
            log.warning(
                'redis connection error %s %s %s, %d retries remaining...',
                addr,
                e.__class__.__name__,
                e,
                settings.conn_retries - retry,
            )
            await asyncio.sleep(settings.conn_retry_delay)
        else:
            raise
    else:
        if retry > 0:
            log.info('redis connection successful')
        return pool

    # recursively attempt to create the pool outside the except block to avoid
    # "During handling of the above exception..." madness
    return await create_pool(
        settings, retry=retry + 1,
        job_serializer=job_serializer, job_deserializer=job_deserializer,
    )


async def log_redis_info(
        redis: ArqRedis, log_func: t.Callable[[str], None],
) -> None:
    with await redis as r:
        info, key_count = await asyncio.gather(r.info(), r.dbsize())
    log_func(
        f'redis_version={info["server"]["redis_version"]} '
        f'mem_usage={info["memory"]["used_memory_human"]} '
        f'clients_connected={info["clients"]["connected_clients"]} '
        f'db_keys={key_count}',
    )



import asyncio
import typing as t
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta

from .types import CoroutineType
from .types import DarqTask
from .utils import get_function_name
from .utils import import_string
from .utils import SecondsTimedelta
from .utils import to_seconds

CronIterable = t.Union[t.Set[int], t.List[int], t.Tuple[int]]


class D:
    month = 'month'
    day = 'day'
    weekday = 'weekday'
    hour = 'hour'
    minute = 'minute'
    second = 'second'
    microsecond = 'microsecond'


dt_fields = D.month, D.day, D.weekday, D.hour, D.minute, D.second, D.microsecond
weekdays = 'mon', 'tues', 'wed', 'thurs', 'fri', 'sat', 'sun'


def _get_next_dt(
        dt_: datetime, options: t.Dict[str, t.Any],
) -> t.Optional[datetime]:
    for field in dt_fields:
        v = options[field]
        if v is None:
            continue
        if field == D.weekday:
            next_v = dt_.weekday()
        else:
            next_v = getattr(dt_, field)
        if isinstance(v, int):
            mismatch = next_v != v
        else:
            assert isinstance(v, (set, list, tuple))
            mismatch = next_v not in v

        if mismatch:
            micro = max(dt_.microsecond - options[D.microsecond], 0)
            if field == D.month:
                if dt_.month == 12:
                    return datetime(dt_.year + 1, 1, 1)
                else:
                    return datetime(dt_.year, dt_.month + 1, 1)
            elif field in (D.day, D.weekday):
                return (
                    dt_
                    + timedelta(days=1)
                    - timedelta(
                        hours=dt_.hour, minutes=dt_.minute, seconds=dt_.second,
                        microseconds=micro,
                    )
                )
            elif field == D.hour:
                return (
                    dt_
                    + timedelta(hours=1)
                    - timedelta(
                        minutes=dt_.minute, seconds=dt_.second,
                        microseconds=micro,
                    )
                )
            elif field == D.minute:
                return (
                    dt_
                    + timedelta(minutes=1)
                    - timedelta(seconds=dt_.second, microseconds=micro)
                )
            elif field == D.second:
                return (
                    dt_
                    + timedelta(seconds=1)
                    - timedelta(microseconds=micro)
                )
            else:
                assert field == D.microsecond, field
                return dt_ + timedelta(
                    microseconds=options['microsecond'] - dt_.microsecond,
                )
    return None


def next_cron(
    previous_dt: datetime,
    *,
    month: t.Union[None, CronIterable, int] = None,
    day: t.Union[None, CronIterable, int] = None,
    weekday: t.Union[None, CronIterable, int, str] = None,
    hour: t.Union[None, CronIterable, int] = None,
    minute: t.Union[None, CronIterable, int] = None,
    second: t.Union[None, CronIterable, int] = 0,
    microsecond: int = 123_456,
) -> datetime:
    """
    Find the next datetime matching the given parameters.
    """
    dt = previous_dt + timedelta(seconds=1)
    if isinstance(weekday, str):
        weekday = weekdays.index(weekday.lower())
    options = dict(
        month=month, day=day, weekday=weekday, hour=hour, minute=minute,
        second=second, microsecond=microsecond,
    )

    while True:
        next_dt = _get_next_dt(dt, options)
        if next_dt is None:
            return dt
        dt = next_dt


@dataclass
class CronJob:
    name: str
    task: DarqTask[CoroutineType]
    month: t.Union[None, CronIterable, int]
    day: t.Union[None, CronIterable, int]
    weekday: t.Union[None, CronIterable, int, str]
    hour: t.Union[None, CronIterable, int]
    minute: t.Union[None, CronIterable, int]
    second: t.Union[None, CronIterable, int]
    microsecond: int
    run_at_startup: bool
    unique: bool
    timeout_s: t.Optional[float]
    keep_result_s: t.Optional[float]
    max_tries: t.Optional[int]
    next_run: t.Optional[datetime] = None

    def set_next(self, dt: datetime) -> None:
        self.next_run = next_cron(
            dt,
            month=self.month,
            day=self.day,
            weekday=self.weekday,
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
        )

    def __repr__(self) -> str:
        return '<CronJob {}>'.format(
            ' '.join(f'{k}={v}' for k, v in self.__dict__.items()),
        )

    def __str__(self) -> str:
        return get_function_name(self.task)


def cron(
    task: t.Union[str, DarqTask[CoroutineType]],
    *,
    name: t.Optional[str] = None,
    month: t.Union[None, CronIterable, int] = None,
    day: t.Union[None, CronIterable, int] = None,
    weekday: t.Union[None, CronIterable, int, str] = None,
    hour: t.Union[None, CronIterable, int] = None,
    minute: t.Union[None, CronIterable, int] = None,
    second: t.Union[None, CronIterable, int] = 0,
    microsecond: int = 123_456,
    run_at_startup: bool = False,
    unique: bool = True,
    timeout: t.Optional[SecondsTimedelta] = None,
    keep_result: t.Optional[float] = 0,
    max_tries: t.Optional[int] = 1,
) -> CronJob:
    """
    Create a cron job, eg. it should be executed at specific times.

    Workers will enqueue this job at or just after the set times.
    If ``unique`` is true (the default) the job will only be run once even
    if multiple workers are running.

    :param task: task function to run
    :param name: name of the job, if None, the name of the task is used
    :param month: month(s) to run the job on, 1 - 12
    :param day: day(s) to run the job on, 1 - 31
    :param weekday: week day(s) to run the job on, 0 - 6 or mon - sun
    :param hour: hour(s) to run the job on, 0 - 23
    :param minute: minute(s) to run the job on, 0 - 59
    :param second: second(s) to run the job on, 0 - 59
    :param microsecond: microsecond(s) to run the job on, defaults to 123456
        as the world is busier at the top of a second, 0 - 1e6
    :param run_at_startup: whether to run as worker starts
    :param unique: whether the job should be only be executed once at each time
    :param timeout: job timeout
    :param keep_result: how long to keep the result for
    :param max_tries: maximum number of tries for the job
    """

    if isinstance(task, str):
        name = name or 'cron:' + task
        task = import_string(task)

    task = t.cast(DarqTask[CoroutineType], task)

    assert asyncio.iscoroutinefunction(task), \
        f'{task} is not a coroutine function'
    timeout = to_seconds(timeout)
    keep_result = to_seconds(keep_result)

    return CronJob(
        name or 'cron:' + task.__qualname__,
        task,
        month,
        day,
        weekday,
        hour,
        minute,
        second,
        microsecond,
        run_at_startup,
        unique,
        timeout,
        keep_result,
        max_tries,
    )



import asyncio
import inspect
import typing as t
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from time import time
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Union

from .types import AnyCallable
from .types import DarqTask


epoch = datetime(1970, 1, 1)
epoch_tz = epoch.replace(tzinfo=timezone.utc)
SecondsTimedelta = Union[int, float, timedelta]


def import_string(dotted_path: str) -> Any:
    """
    Stolen approximately from Django. Import a dotted module path and return
    the attribute/class designated by the last name in the path.
    Raise ImportError if the import fails.
    """
    from importlib import import_module

    try:
        module_path, class_name = dotted_path.strip(' ').rsplit('.', 1)
    except ValueError as e:
        raise ImportError(
            f'"{dotted_path}" doesn\'t look like a module path') from e

    module = import_module(module_path)
    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise ImportError(
            f'Module "{module_path}" does not define a "{class_name}" '
            f'attribute') from e


def as_int(f: float) -> int:
    return int(round(f))


def timestamp_ms() -> int:
    return as_int(time() * 1000)


def to_unix_ms(dt: datetime) -> int:
    """
    convert a datetime to number of milliseconds since 1970 and calculate
    timezone offset
    """
    utcoffset = dt.utcoffset()
    ep = epoch if utcoffset is None else epoch_tz
    return as_int((dt - ep).total_seconds() * 1000)


def ms_to_datetime(unix_ms: int) -> datetime:
    return epoch + timedelta(seconds=unix_ms / 1000)


def to_ms(td: Optional[SecondsTimedelta]) -> Optional[int]:
    if td is None:
        return td
    elif isinstance(td, timedelta):
        td = td.total_seconds()
    return as_int(td * 1000)


def to_seconds_strict(td: SecondsTimedelta) -> float:
    if isinstance(td, timedelta):
        return td.total_seconds()
    return td


def to_seconds(td: Optional[SecondsTimedelta]) -> Optional[float]:
    if td is None:
        return td
    return to_seconds_strict(td)


DEFAULT_CURTAIL = 80


def truncate(s: str, length: int = DEFAULT_CURTAIL) -> str:
    """
    Truncate a string and add an ellipsis (three dots) to the end
    if it was too long

    :param s: string to possibly truncate
    :param length: length to truncate the string to
    """
    if len(s) > length:
        s = s[: length - 1] + '…'
    return s


def args_to_string(args: Sequence[Any], kwargs: Mapping[Any, Any]) -> str:
    arguments = ''
    if args:
        arguments = ', '.join(map(repr, args))
    if kwargs:
        if arguments:
            arguments += ', '
        arguments += ', '.join(f'{k}={v!r}' for k, v in sorted(kwargs.items()))
    return truncate(arguments)


def get_function_name(func: t.Union[AnyCallable, DarqTask[AnyCallable]]) -> str:
    module_str = func.__module__
    if module_str == '__main__':
        module = inspect.getmodule(func)
        if module and module.__spec__:
            module_str = module.__spec__.name

    return f'{module_str}.{func.__name__}'


async def poll(
        step: float, *, timeout: t.Optional[float] = None,
) -> t.AsyncIterator[float]:
    loop = asyncio.get_event_loop()
    start = loop.time()
    while True:
        before = loop.time()
        yield before - start
        after = loop.time()
        if timeout and after - start >= timeout:
            return
        wait = max([0, step - after + before])
        await asyncio.sleep(wait)



import asyncio
import pickle

import pytest

from darq.constants import default_queue_name
from darq.constants import in_progress_key_prefix
from darq.constants import job_key_prefix
from darq.constants import result_key_prefix
from darq.jobs import DeserializationError
from darq.jobs import deserialize_job_raw
from darq.jobs import Job
from darq.jobs import JobResult
from darq.jobs import JobStatus
from darq.jobs import serialize_result
from .utils import CloseToNow


async def test_job_in_progress(arq_redis):
    await arq_redis.set(in_progress_key_prefix + 'foobar', b'1')
    j = Job('foobar', arq_redis)
    assert JobStatus.in_progress == await j.status()
    assert str(j) == '<darq job foobar>'


async def test_unknown(arq_redis):
    j = Job('foobar', arq_redis)
    assert JobStatus.not_found == await j.status()
    info = await j.info()
    assert info is None


async def test_result_timeout(arq_redis):
    j = Job('foobar', arq_redis)
    with pytest.raises(asyncio.TimeoutError):
        await j.result(0.1, pole_delay=0)


async def foobar(*args, **kwargs):
    return 42


async def test_enqueue_job(
        darq, arq_redis, worker_factory, queue_name=default_queue_name,
):
    darq.task(foobar)

    j = await arq_redis.enqueue_job(
        'tests.test_jobs.foobar', [1, 2], {'c': 3}, queue_name=queue_name,
    )
    assert isinstance(j, Job)
    assert JobStatus.queued == await j.status()
    worker = worker_factory(darq, queue_name=queue_name)
    await worker.main()
    r = await j.result(pole_delay=0)
    assert r == 42
    assert JobStatus.complete == await j.status()
    info = await j.info()
    assert info == JobResult(
        job_try=1,
        function='tests.test_jobs.foobar',
        args=[1, 2],
        kwargs={'c': 3},
        enqueue_time=CloseToNow(),
        success=True,
        result=42,
        start_time=CloseToNow(),
        finish_time=CloseToNow(),
        score=None,
    )
    results = await arq_redis.all_job_results()
    assert results == [
        JobResult(
            function='tests.test_jobs.foobar',
            args=[1, 2],
            kwargs={'c': 3},
            job_try=1,
            enqueue_time=CloseToNow(),
            success=True,
            result=42,
            start_time=CloseToNow(),
            finish_time=CloseToNow(),
            score=None,
            job_id=j.job_id,
        ),
    ]


async def test_enqueue_job_alt_queue(darq, arq_redis, worker_factory):
    await test_enqueue_job(
        darq, arq_redis, worker_factory, queue_name='custom_queue',
    )


async def test_cant_unpickle_at_all():
    class Foobar:
        def __getstate__(self):
            raise TypeError("this doesn't pickle")

    r1 = serialize_result(
        'foobar', (1,), {}, 1, 123, True, Foobar(), 123, 123, 'testing',
    )
    assert isinstance(r1, bytes)
    r2 = serialize_result(
        'foobar', (Foobar(),), {}, 1, 123, True, Foobar(), 123, 123, 'testing',
    )
    assert r2 is None


async def test_custom_serializer():
    class Foobar:
        def __getstate__(self):
            raise TypeError("this doesn't pickle")

    def custom_serializer(x):
        return b'0123456789'

    r1 = serialize_result(
        'foobar', (1,), {}, 1, 123, True, Foobar(), 123, 123, 'testing',
        serializer=custom_serializer,
    )
    assert r1 == b'0123456789'
    r2 = serialize_result(
        'foobar', (Foobar(),), {}, 1, 123, True, Foobar(), 123, 123, 'testing',
        serializer=custom_serializer,
    )
    assert r2 == b'0123456789'


async def foobar_sum(a, b):
    return a + b


async def test_deserialize_result(darq, arq_redis, worker_factory):
    darq.task(foobar_sum)

    j = await arq_redis.enqueue_job('tests.test_jobs.foobar_sum', [1, 2], {})
    assert JobStatus.queued == await j.status()
    worker = worker_factory(darq)
    await worker.run_check()
    assert await j.result(pole_delay=0) == 3
    assert await j.result(pole_delay=0) == 3
    info = await j.info()
    assert info.args == [1, 2]
    await arq_redis.set(result_key_prefix + j.job_id, b'invalid pickle data')
    with pytest.raises(
            DeserializationError, match='unable to deserialize job result',
    ):
        assert await j.result(pole_delay=0) == 3


async def test_deserialize_info(arq_redis):
    j = await arq_redis.enqueue_job('foobar', [1, 2], {})
    assert JobStatus.queued == await j.status()
    await arq_redis.set(job_key_prefix + j.job_id, b'invalid pickle data')

    with pytest.raises(DeserializationError, match='unable to deserialize job'):
        assert await j.info()


async def test_deserialize_job_raw():
    serialized = pickle.dumps({'f': 1, 'a': 2, 'k': 3, 't': 4, 'et': 5})
    assert deserialize_job_raw(serialized) == (1, 2, 3, 4, 5)
    with pytest.raises(DeserializationError, match='unable to deserialize job'):
        deserialize_job_raw(b'123')



import typing as t
from collections import UserDict

from . import worker

if t.TYPE_CHECKING:  # pragma: no cover
    BaseRegistry = UserDict[str, worker.Task]
else:
    BaseRegistry = UserDict


class Registry(BaseRegistry):

    def add(self, task: 'worker.Task') -> None:
        self[task.name] = task

    def get_functions(self) -> t.Sequence['worker.Task']:
        return tuple(self.values())

    def get_function_names(self) -> t.Sequence[str]:
        return tuple(task.name for task in self.values())



import logging
import re
from datetime import timedelta

import pytest

import darq
from darq import utils
from darq.connections import create_pool
from darq.connections import log_redis_info
from darq.connections import RedisSettings
from . import redis_settings


def test_settings_changed():
    settings = RedisSettings(port=123)
    assert settings.port == 123
    assert (
        '<RedisSettings host=localhost port=123 database=0 password=None '
        'ssl=None conn_timeout=1 conn_retries=5 conn_retry_delay=1 '
        'sentinel=False sentinel_master=mymaster sentinel_timeout=0.2>'
    ) == str(settings)


async def test_redis_timeout(mocker):
    mocker.spy(darq.connections.asyncio, 'sleep')
    with pytest.raises(OSError):
        await create_pool(RedisSettings(port=0, conn_retry_delay=0))
    assert darq.connections.asyncio.sleep.call_count == 5


async def test_redis_sentinel_failure():
    settings = RedisSettings()
    settings.host = [(redis_settings.host, 6379), (redis_settings.host, 6379)]
    settings.sentinel = True
    try:
        pool = await create_pool(settings)
        await pool.ping('ping')
    except Exception as e:
        assert (
            'unknown command `SENTINEL`' in str(e)
            or 'Could not connect to any sentinel' in str(e)
        )


async def test_redis_success_log(caplog):
    caplog.set_level(logging.INFO)
    pool = await create_pool(redis_settings)
    assert 'redis connection successful' not in [
        r.message for r in caplog.records
    ]
    pool.close()
    await pool.wait_closed()

    pool = await create_pool(redis_settings, retry=1)
    assert 'redis connection successful' in [r.message for r in caplog.records]
    pool.close()
    await pool.wait_closed()


async def test_redis_log():
    redis = await create_pool(redis_settings)
    await redis.flushall()
    await redis.set(b'a', b'1')
    await redis.set(b'b', b'2')

    log_msgs = []

    def _log(s):
        log_msgs.append(s)

    await log_redis_info(redis, _log)
    assert len(log_msgs) == 1
    assert re.search(r'redis_version=\d\.', log_msgs[0]), log_msgs
    assert log_msgs[0].endswith(' db_keys=2')


def test_truncate():
    assert utils.truncate('123456', 4) == '123…'


def test_args_to_string():
    assert utils.args_to_string((), {'d': 4}) == 'd=4'
    assert utils.args_to_string((1, 2, 3), {}) == '1, 2, 3'
    assert utils.args_to_string((1, 2, 3), {'d': 4}) == '1, 2, 3, d=4'


@pytest.mark.parametrize('input_,output', [
    (timedelta(days=1), 86_400_000),
    (42, 42000),
    (42.123, 42123),
    (42.123_987, 42124),
    (None, None),
])
def test_to_ms(input_, output):
    assert utils.to_ms(input_) == output


@pytest.mark.parametrize('input_,output', [
    (timedelta(days=1), 86400),
    (42, 42),
    (42.123, 42.123),
    (None, None),
])
def test_to_seconds(input_, output):
    assert utils.to_seconds(input_) == output



import asyncio
import dataclasses
import logging
import signal
import typing as t
from datetime import datetime
from functools import partial
from signal import Signals
from time import time

import async_timeout
from aioredis import MultiExecError

from .connections import create_pool
from .connections import log_redis_info
from .connections import RedisSettings
from .constants import default_queue_name
from .constants import health_check_key_suffix
from .constants import in_progress_key_prefix
from .constants import job_key_prefix
from .constants import result_key_prefix
from .constants import retry_key_prefix
from .jobs import deserialize_job_raw
from .jobs import Deserializer
from .jobs import JobResult
from .jobs import SerializationError
from .jobs import serialize_result
from .jobs import Serializer
from .types import CoroutineType
from .utils import args_to_string
from .utils import ms_to_datetime
from .utils import poll
from .utils import SecondsTimedelta
from .utils import timestamp_ms
from .utils import to_ms
from .utils import to_seconds
from .utils import to_seconds_strict
from .utils import truncate

if t.TYPE_CHECKING:  # pragma: no cover
    from darq import Darq
    from .connections import ArqRedis

log = logging.getLogger('darq.worker')
no_result = object()

CtxType = t.Dict[t.Any, t.Any]


class Task(t.NamedTuple):
    name: str
    coroutine: CoroutineType
    timeout_s: t.Optional[float]
    keep_result_s: t.Optional[float]
    max_tries: t.Optional[int]
    with_ctx: bool

    @classmethod
    def new(
        cls,
        name: str,
        coroutine: CoroutineType,
        timeout: t.Optional[SecondsTimedelta] = None,
        keep_result: t.Optional[SecondsTimedelta] = None,
        max_tries: t.Optional[int] = None,
        with_ctx: bool = False,
    ) -> 'Task':
        return cls(
            name=name, coroutine=coroutine, timeout_s=to_seconds(timeout),
            keep_result_s=to_seconds(keep_result), max_tries=max_tries,
            with_ctx=with_ctx,
        )


class Retry(RuntimeError):
    """
    Special exception to retry the job (if ``max_retries`` hasn't been reached).

    :param defer: duration to wait before rerunning the job
    """

    def __init__(self, defer: t.Optional[SecondsTimedelta] = None) -> None:
        self.defer_score = to_ms(defer)

    def __repr__(self) -> str:
        return f'<Retry defer {(self.defer_score or 0) / 1000:0.2f}s>'

    def __str__(self) -> str:
        return repr(self)


class JobExecutionFailed(RuntimeError):
    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, JobExecutionFailed):
            return self.args == other.args
        return False


class FailedJobs(RuntimeError):
    def __init__(self, count: int, job_results: t.Sequence[JobResult]) -> None:
        self.count = count
        self.job_results = job_results

    def __str__(self) -> str:
        if self.count == 1 and self.job_results:
            exc = self.job_results[0].result
            return f'1 job failed {exc!r}'
        else:
            return (
                f'{self.count} jobs failed:\n'
                + '\n'.join(repr(r.result) for r in self.job_results)
            )

    def __repr__(self) -> str:
        return f'<{str(self)}>'


@dataclasses.dataclass
class WorkerSettings:
    """
    Worker settings

    :param queue_name: queue name to get jobs from
    :param burst: whether to stop the worker once all jobs have been run
    :param max_jobs: maximum number of jobs to run at a time
    :param job_timeout: default job timeout (max run time)
    :param keep_result: default duration to keep job results for
    :param poll_delay: duration between polling the queue for new jobs
    :param queue_read_limit: the maximum number of jobs to pull from the queue
                             each time it's polled;
                             by default it equals ``max_jobs``
    :param max_tries: default maximum number of times to retry a job
    :param health_check_interval: how often to set the health check key
    :param health_check_key: redis key under which health check is set
    :param retry_jobs: whether to retry jobs on Retry or CancelledError or not
    :param max_burst_jobs: the maximum number of jobs to process in burst mode
                           (disabled with negative values)
    :param job_serializer: a function that serializes Python objects to bytes,
                           defaults to pickle.dumps
    :param job_deserializer: a function that deserializes bytes into Python
                             objects, defaults to pickle.loads
    """
    queue_name: str = default_queue_name
    burst: bool = False
    max_jobs: int = 10
    job_timeout: SecondsTimedelta = 300
    keep_result: SecondsTimedelta = 3600
    poll_delay: SecondsTimedelta = 0.5
    queue_read_limit: t.Optional[int] = None
    max_tries: int = 5
    health_check_interval: SecondsTimedelta = 3600
    health_check_key: t.Optional[str] = None
    retry_jobs: bool = True
    max_burst_jobs: int = -1
    job_serializer: t.Optional[Serializer] = None
    job_deserializer: t.Optional[Deserializer] = None


class Worker:
    """
    Main class for running jobs.

    :param app: instance of :func:`darq.app.Darq`
    :param worker_settings: instance of :class:`darq.worker.WorkerSettings`
    """

    def __init__(
            self, app: 'Darq', **replace_kwargs: t.Dict[str, t.Any],
    ) -> None:
        settings = dataclasses.replace(app.worker_settings, **replace_kwargs)
        self.app = app
        self.functions: t.Dict[str, Task] = dict(app.registry)
        self.queue_name = settings.queue_name

        assert len(self.functions) > 0, \
            'at least one function or cron_job must be registered'
        self.burst = settings.burst
        self.on_startup = app.on_startup
        self.on_shutdown = app.on_shutdown
        self.on_job_prerun = app.on_job_prerun
        self.on_job_postrun = app.on_job_postrun
        self.sem = asyncio.BoundedSemaphore(settings.max_jobs)
        self.job_timeout_s = to_seconds_strict(settings.job_timeout)
        self.keep_result_s = to_seconds_strict(settings.keep_result)
        self.poll_delay_s = to_seconds_strict(settings.poll_delay)
        self.queue_read_limit = (
            settings.queue_read_limit or settings.max_jobs
        )
        self._queue_read_offset = 0
        self.max_tries = settings.max_tries
        self.health_check_interval = to_seconds_strict(
            settings.health_check_interval,
        )
        if settings.health_check_key is None:
            self.health_check_key = self.queue_name + health_check_key_suffix
        else:
            self.health_check_key = settings.health_check_key

        self.pool: 'ArqRedis' = None  # type: ignore[assignment]
        self.redis_settings = app.redis_settings or RedisSettings()

        self.tasks: t.List[asyncio.Task[t.Any]] = []
        self.main_task: t.Optional[asyncio.Task[None]] = None
        self.graceful_shutdown_task: t.Optional[asyncio.Task[None]] = None
        self.graceful_shutdown_timeout = 30
        self.loop = asyncio.get_event_loop()
        self.ctx = app.ctx or {}
        max_timeout = max(
            f.timeout_s or self.job_timeout_s for f in self.functions.values()
        )
        self.in_progress_timeout_s = max_timeout + 10
        self.jobs_complete = 0
        self.jobs_retried = 0
        self.jobs_failed = 0
        self._last_health_check: float = 0
        self._last_health_check_log: t.Optional[str] = None
        self._add_signal_handler(signal.SIGINT, self.handle_sig)
        self._add_signal_handler(signal.SIGTERM, self.handle_sig)
        self.on_stop: t.Optional[t.Callable[[Signals], None]] = None
        # whether or not to retry jobs on Retry and CancelledError
        self.retry_jobs = settings.retry_jobs
        self.max_burst_jobs = settings.max_burst_jobs
        self.job_serializer = settings.job_serializer
        self.job_deserializer = settings.job_deserializer

    def run(self) -> None:
        """
        Sync function to run the worker, finally closes worker connections.
        """
        self.main_task = self.loop.create_task(self.main())
        try:
            self.loop.run_until_complete(self.main_task)
        except asyncio.CancelledError:  # pragma: no cover
            # happens on shutdown, fine
            pass
        finally:
            self.loop.run_until_complete(self.close())

    async def async_run(self) -> None:
        """
        Asynchronously run the worker, does not close connections.
        Useful when testing.
        """
        self.main_task = self.loop.create_task(self.main())
        await self.main_task

    async def run_check(
            self,
            retry_jobs: t.Optional[bool] = None,
            max_burst_jobs: t.Optional[int] = None,
    ) -> int:
        """
        Run :func:`darq.worker.Worker.async_run`, check for failed jobs
        and raise :class:`darq.worker.FailedJobs` if any jobs have failed.

        :return: number of completed jobs
        """
        if retry_jobs is not None:
            self.retry_jobs = retry_jobs
        if max_burst_jobs is not None:
            self.max_burst_jobs = max_burst_jobs
        await self.async_run()
        if self.jobs_failed:
            failed_job_results = [
                r for r in await self.pool.all_job_results()
                if not r.success
            ]
            raise FailedJobs(self.jobs_failed, failed_job_results)
        else:
            return self.jobs_complete

    async def main(self) -> None:
        log.info(
            'Starting worker for %d functions: %s',
            len(self.functions), ', '.join(self.functions),
        )
        self.pool = await create_pool(self.redis_settings)
        await log_redis_info(self.pool, log.info)
        self.ctx['redis'] = self.pool
        await self.app.connect(self.pool)
        if self.on_startup:
            await self.on_startup(self.ctx)

        async for _ in poll(self.poll_delay_s):
            if self.graceful_shutdown_task:
                self.clean_tasks_done()
                continue

            await self._poll_iteration()

            if self.burst:
                if 0 <= self.max_burst_jobs <= self._jobs_started():
                    await asyncio.gather(*self.tasks)
                    return
                queued_jobs = await self.pool.zcard(self.queue_name)
                if queued_jobs == 0:
                    await asyncio.gather(*self.tasks)
                    return

    async def _poll_iteration(self) -> None:
        count = self.queue_read_limit
        if self.burst and self.max_burst_jobs >= 0:
            burst_jobs_remaining = self.max_burst_jobs - self._jobs_started()
            if burst_jobs_remaining < 1:
                return
            count = min(burst_jobs_remaining, count)

        async with self.sem:
            # don't bother with zrangebyscore until we have "space"
            # to run the jobs
            now = timestamp_ms()
            job_ids = await self.pool.zrangebyscore(
                self.queue_name, offset=self._queue_read_offset,
                count=count, max=now,
            )
        await self.run_jobs(job_ids)
        self.clean_tasks_done()
        await self.heart_beat()

    def clean_tasks_done(self) -> None:
        for task in self.tasks:
            if task.done():
                self.tasks.remove(task)
                # required to make sure errors in run_job get propagated
                task.result()

    async def run_jobs(self, job_ids: t.Sequence[str]) -> None:
        for job_id in job_ids:
            await self.sem.acquire()
            in_progress_key = in_progress_key_prefix + job_id
            with await self.pool as conn:
                _, _, ongoing_exists, score = await asyncio.gather(
                    conn.unwatch(),
                    conn.watch(in_progress_key),
                    conn.exists(in_progress_key),
                    conn.zscore(self.queue_name, job_id),
                )
                if ongoing_exists or not score:
                    # job already started elsewhere, or already finished and
                    # removed from queue
                    self.sem.release()
                    continue

                tr = conn.multi_exec()
                tr.setex(in_progress_key, self.in_progress_timeout_s, b'1')
                try:
                    await tr.execute()
                except MultiExecError:  # pragma: no cover
                    # job already started elsewhere since we got 'existing'
                    self.sem.release()
                    log.debug(
                        'multi-exec error, job %s already started elsewhere',
                        job_id,
                    )
                    # https://github.com/samuelcolvin/arq/issues/131,
                    # avoid warnings in log
                    await asyncio.gather(*tr._results, return_exceptions=True)
                else:
                    task = self.loop.create_task(self.run_job(job_id, score))
                    task.add_done_callback(lambda _: self.sem.release())
                    self.tasks.append(task)

    async def run_job(self, job_id: str, score: float) -> None:
        start_ms = timestamp_ms()
        v, job_try, _ = await asyncio.gather(
            self.pool.get(job_key_prefix + job_id, encoding=None),
            self.pool.incr(retry_key_prefix + job_id),
            self.pool.expire(retry_key_prefix + job_id, 88400),
        )
        kwargs: t.Dict[str, t.Any]
        args: t.Sequence[t.Any]
        function_name, args, kwargs, enqueue_time_ms = '<unknown>', (), {}, 0

        async def job_failed(exc: Exception) -> None:
            self.jobs_failed += 1
            result_data_ = serialize_result(
                function=function_name,
                args=args,
                kwargs=kwargs,
                job_try=job_try,
                enqueue_time_ms=enqueue_time_ms,
                success=False,
                result=exc,
                start_ms=start_ms,
                finished_ms=timestamp_ms(),
                ref=f'{job_id}:{function_name}',
                serializer=self.job_serializer,
            )
            await asyncio.shield(self.abort_job(job_id, result_data_))

        if not v:
            log.warning('job %s expired', job_id)
            await job_failed(JobExecutionFailed('job expired'))
            return

        try:
            (
                function_name, args, kwargs, enqueue_job_try, enqueue_time_ms,
            ) = deserialize_job_raw(v, deserializer=self.job_deserializer)
        except SerializationError as e:
            log.exception('deserializing job %s failed', job_id)
            await job_failed(e)
            return

        try:
            function = self.functions[function_name]
        except KeyError:
            log.error('job %s, function %r not found', job_id, function_name)
            await job_failed(
                JobExecutionFailed(f'function {function_name!r} not found'),
            )
            return

        ref = f'{job_id}:{function_name}'

        if enqueue_job_try and enqueue_job_try > job_try:
            job_try = enqueue_job_try
            await self.pool.setex(
                retry_key_prefix + job_id, 88400, str(job_try),
            )

        max_tries = (
            self.max_tries
            if function.max_tries is None
            else function.max_tries
        )
        if job_try > max_tries:
            log.warning(
                '%6.2fs ! %s max retries %d exceeded',
                (timestamp_ms() - enqueue_time_ms) / 1000, ref, max_tries,
            )
            self.jobs_failed += 1
            result_data = serialize_result(
                function_name,
                args,
                kwargs,
                job_try,
                enqueue_time_ms,
                False,
                JobExecutionFailed(f'max {max_tries} retries exceeded'),
                start_ms,
                timestamp_ms(),
                ref,
                serializer=self.job_serializer,
            )
            return await asyncio.shield(self.abort_job(job_id, result_data))

        result = no_result
        exc_extra = None
        finish = False
        timeout_s = (
            self.job_timeout_s
            if function.timeout_s is None
            else function.timeout_s
        )
        incr_score: t.Optional[float] = None
        job_ctx = {
            'job_id': job_id,
            'job_try': job_try,
            'enqueue_time': ms_to_datetime(enqueue_time_ms),
            'score': score,
            'metadata': kwargs.pop('__metadata__', {}),
        }
        ctx = {**self.ctx, **job_ctx}
        start_ms = timestamp_ms()
        success = False
        try:
            if self.on_job_prerun:
                await self.on_job_prerun(ctx, function, args, kwargs)

            s = args_to_string(args, kwargs)
            extra = f' try={job_try}' if job_try > 1 else ''
            if (start_ms - score) > 1200:
                extra += f' delayed={(start_ms - score) / 1000:0.2f}s'
            log.info(
                '%6.2fs → %s(%s)%s',
                (start_ms - enqueue_time_ms) / 1000, ref, s, extra,
            )

            coro = function.coroutine
            if function.with_ctx:
                coro = partial(coro, ctx)

            # run repr(result) and extra inside try/except as they can
            # raise exceptions
            try:
                async with async_timeout.timeout(timeout_s):
                    result = await coro(*args, **kwargs)
            except Exception as e:
                exc_extra = getattr(e, 'extra', None)
                if callable(exc_extra):
                    exc_extra = exc_extra()
                raise
            else:
                result_str = '' if result is None else truncate(repr(result))
        except Exception as e:
            finished_ms = timestamp_ms()
            t_ = (finished_ms - start_ms) / 1000
            if self.retry_jobs and isinstance(e, Retry):
                incr_score = e.defer_score
                log.info(
                    '%6.2fs ↻ %s retrying job in %0.2fs',
                    t_, ref, (e.defer_score or 0) / 1000,
                )
                if e.defer_score:
                    incr_score = e.defer_score + (timestamp_ms() - score)
                self.jobs_retried += 1
            elif self.retry_jobs and isinstance(e, asyncio.CancelledError):
                log.info('%6.2fs ↻ %s cancelled, will be run again', t_, ref)
                self.jobs_retried += 1
            else:
                log.exception(
                    '%6.2fs ! %s failed, %s: %s',
                    t_, ref, e.__class__.__name__, e,
                    extra={'extra': exc_extra},
                )
                result = e
                finish = True
                self.jobs_failed += 1
        else:
            success = True
            finished_ms = timestamp_ms()
            log.info(
                '%6.2fs ← %s ● %s',
                (finished_ms - start_ms) / 1000, ref, result_str,
            )
            finish = True
            self.jobs_complete += 1

        result_timeout_s = (
            self.keep_result_s
            if function.keep_result_s is None
            else function.keep_result_s
        )
        result_data = None
        if result is not no_result and result_timeout_s > 0:
            result_data = serialize_result(
                function_name,
                args,
                kwargs,
                job_try,
                enqueue_time_ms,
                success,
                result,
                start_ms,
                finished_ms,
                ref,
                serializer=self.job_serializer,
            )

        await asyncio.shield(self.finish_job(
            job_id, finish, result_data, result_timeout_s, incr_score,
        ))
        if self.on_job_postrun:
            await self.on_job_postrun(ctx, function, args, kwargs, result)

    async def finish_job(
            self, job_id: str, finish: bool, result_data: t.Optional[bytes],
            result_timeout_s: t.Optional[float], incr_score: t.Optional[float],
    ) -> None:
        with await self.pool as conn:
            await conn.unwatch()
            tr = conn.multi_exec()
            delete_keys = [in_progress_key_prefix + job_id]
            if finish:
                if result_data:
                    tr.setex(
                        result_key_prefix + job_id, result_timeout_s,
                        result_data,
                    )
                delete_keys += [
                    retry_key_prefix + job_id, job_key_prefix + job_id,
                ]
                tr.zrem(self.queue_name, job_id)
            elif incr_score:
                tr.zincrby(self.queue_name, incr_score, job_id)
            tr.delete(*delete_keys)
            await tr.execute()

    async def abort_job(
            self, job_id: str, result_data: t.Optional[bytes],
    ) -> None:
        with await self.pool as conn:
            await conn.unwatch()
            tr = conn.multi_exec()
            tr.delete(
                retry_key_prefix + job_id,
                in_progress_key_prefix + job_id,
                job_key_prefix + job_id,
            )
            tr.zrem(self.queue_name, job_id)
            # result_data would only be None if serializing the result fails
            if (
                    result_data is not None
                    and self.keep_result_s > 0
            ):  # pragma: no branch
                tr.setex(
                    result_key_prefix + job_id, self.keep_result_s, result_data,
                )
            await tr.execute()

    async def heart_beat(self) -> None:
        await self.record_health()

    async def record_health(self) -> None:
        now_ts = time()
        if (now_ts - self._last_health_check) < self.health_check_interval:
            return
        self._last_health_check = now_ts
        pending_tasks = self.running_job_count()
        queued = await self.pool.zcard(self.queue_name)
        info = (
            f'{datetime.now():%b-%d %H:%M:%S} j_complete={self.jobs_complete} '
            f'j_failed={self.jobs_failed} j_retried={self.jobs_retried} '
            f'j_ongoing={pending_tasks} queued={queued}'
        )
        await self.pool.setex(
            self.health_check_key,
            self.health_check_interval + 1,
            info.encode(),
        )
        log_suffix = info[info.index('j_complete='):]
        if (
                self._last_health_check_log
                and log_suffix != self._last_health_check_log
        ):
            log.info('recording health: %s', info)
            self._last_health_check_log = log_suffix
        elif not self._last_health_check_log:
            self._last_health_check_log = log_suffix

    def _add_signal_handler(
            self, signal: signal.Signals, handler: t.Callable[[int], None],
    ) -> None:
        self.loop.add_signal_handler(signal, partial(handler, signal))

    def _jobs_started(self) -> int:
        return (
            self.jobs_complete
            + self.jobs_retried
            + self.jobs_failed
            + len(self.tasks)
        )

    def running_job_count(self) -> int:
        return sum(not task.done() for task in self.tasks)

    async def graceful_shutdown(self, signum: int) -> None:
        running_job_count = self.running_job_count()
        if running_job_count:
            log.info(
                'Warm shutdown. Awaiting for %d jobs with %d seconds timeout.',
                running_job_count, self.graceful_shutdown_timeout,
            )
            async for _ in poll(0.1, timeout=self.graceful_shutdown_timeout):
                if not self.running_job_count():
                    break
        self.shutdown(signum)

    def shutdown(self, signum: int) -> None:
        sig = Signals(signum)
        log.info(
            'shutdown on %s ◆ %d jobs complete ◆ %d failed ◆ %d retries ◆ '
            '%d ongoing to cancel',
            sig.name,
            self.jobs_complete,
            self.jobs_failed,
            self.jobs_retried,
            len(self.tasks),
        )
        for task in self.tasks:
            if not task.done():
                task.cancel()
        self.main_task and self.main_task.cancel()
        self.on_stop and self.on_stop(sig)

    def handle_sig(self, signum: int) -> None:
        if self.graceful_shutdown_task:
            self.graceful_shutdown_task.cancel()
            self.shutdown(signum)
        else:
            self.graceful_shutdown_task = self.loop.create_task(
                self.graceful_shutdown(signum),
            )

    async def close(self) -> None:
        if not self.pool:
            return
        await asyncio.gather(*self.tasks)
        await self.pool.delete(self.health_check_key)
        if self.on_shutdown:
            await self.on_shutdown(self.ctx)
        await self.app.disconnect()
        self.pool.close()
        await self.pool.wait_closed()
        self.pool = None  # type: ignore[assignment]

    def __repr__(self) -> str:
        return (
            f'<Worker j_complete={self.jobs_complete} '
            f'j_failed={self.jobs_failed} j_retried={self.jobs_retried} '
            f'j_ongoing={self.running_job_count()}>'
        )


async def async_check_health(
        redis_settings: t.Optional[RedisSettings],
        health_check_key: t.Optional[str] = None,
        queue: t.Optional[str] = None,
) -> int:
    redis_settings = redis_settings or RedisSettings()
    redis = await create_pool(redis_settings)
    queue = queue or default_queue_name
    health_check_key = health_check_key or (queue + health_check_key_suffix)

    data = await redis.get(health_check_key)
    if not data:
        log.warning('Health check failed: no health check sentinel value found')
        result = 1
    else:
        log.info('Health check successful: %s', data)
        result = 0
    redis.close()
    await redis.wait_closed()
    return result


def check_health(darq: 'Darq', queue: str) -> int:
    """
    Run a health check on the worker and return the appropriate exit code.
    :return: 0 if successful, 1 if not
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(
        async_check_health(
            darq.redis_settings, darq.worker_settings.health_check_key, queue,
        ),
    )


def create_worker(
        darq: 'Darq', **overwrite_settings: t.Dict[str, t.Any],
) -> Worker:
    return Worker(darq, **overwrite_settings)


def run_worker(
        darq: 'Darq', **overwrite_settings: t.Dict[str, t.Any],
) -> Worker:
    worker = create_worker(darq, **overwrite_settings)
    worker.run()
    return worker



from .app import Darq
from .connections import RedisSettings
from .cron import cron
from .types import JobCtx
from .version import __version__
from .worker import Retry

try:
    from aioredis import __version__ as aioredis_version
except ImportError:
    raise ImportError(
        'Please, install "aioredis" (or "evo-aioredis" - for Python 3.11+)')

AIOREDIS_MAJOR_VERSION = int(aioredis_version.split('.')[0])
if AIOREDIS_MAJOR_VERSION > 1:
    raise ImportError('Darq only supports "aioredis" versions <2.0')

__all__ = ['Darq', 'JobCtx', 'RedisSettings', 'cron', 'Retry', '__version__']



from datetime import datetime
from datetime import timedelta
from random import random

import pytest

from darq.cron import cron
from darq.cron import next_cron


@pytest.mark.parametrize('previous,expected,kwargs', [
    (
        datetime(2016, 6, 1, 12, 10, 10),
        datetime(2016, 6, 1, 12, 10, 20, microsecond=123_456),
        dict(second=20),
    ),
    (
        datetime(2016, 6, 1, 12, 10, 10),
        datetime(2016, 6, 1, 12, 11, 0, microsecond=123_456),
        dict(minute=11),
    ),
    (
        datetime(2016, 6, 1, 12, 10, 10),
        datetime(2016, 6, 1, 12, 10, 20),
        dict(second=20, microsecond=0)),
    (
        datetime(2016, 6, 1, 12, 10, 10),
        datetime(2016, 6, 1, 12, 11, 0),
        dict(minute=11, microsecond=0)),
    (
        datetime(2016, 6, 1, 12, 10, 11),
        datetime(2017, 6, 1, 12, 10, 10, microsecond=123_456),
        dict(month=6, day=1, hour=12, minute=10, second=10),
    ),
    (
        datetime(2016, 6, 1, 12, 10, 10, microsecond=1),
        datetime(2016, 7, 1, 12, 10, 10),
        dict(day=1, hour=12, minute=10, second=10, microsecond=0),
    ),
    (
        datetime(2032, 1, 31, 0, 0, 0),
        datetime(2032, 2, 28, 0, 0, 0, microsecond=123_456),
        dict(day=28),
    ),
    (
        datetime(2032, 1, 1, 0, 5),
        datetime(2032, 1, 1, 4, 0, microsecond=123_456),
        dict(hour=4),
    ),
    (
        datetime(2032, 1, 1, 0, 0),
        datetime(2032, 1, 1, 4, 2, microsecond=123_456),
        dict(hour=4, minute={2, 4, 6}),
    ),
    (
        datetime(2032, 1, 1, 0, 5),
        datetime(2032, 1, 1, 4, 2, microsecond=123_456),
        dict(hour=4, minute={2, 4, 6}),
    ),
    (
        datetime(2032, 2, 5, 0, 0, 0),
        datetime(2032, 3, 31, 0, 0, 0, microsecond=123_456),
        dict(day=31),
    ),
    (
        datetime(2001, 1, 1, 0, 0, 0),  # Monday
        datetime(2001, 1, 7, 0, 0, 0, microsecond=123_456),
        dict(weekday='Sun'),  # Sunday
    ),
    (
        datetime(2001, 1, 1, 0, 0, 0),
        datetime(2001, 1, 7, 0, 0, 0, microsecond=123_456),
        dict(weekday=6),  # Sunday
    ),
    (
        datetime(2001, 1, 1, 0, 0, 0),
        datetime(2001, 11, 7, 0, 0, 0, microsecond=123_456),
        dict(month=11, weekday=2),
    ),
    (
        datetime(2001, 1, 1, 0, 0, 0),
        datetime(2001, 1, 3, 0, 0, 0, microsecond=123_456),
        dict(weekday='wed'),
    ),
])
def test_next_cron(previous, expected, kwargs):
    assert next_cron(previous, **kwargs) == expected


def test_next_cron_invalid():
    with pytest.raises(ValueError):
        next_cron(datetime(2001, 1, 1, 0, 0, 0), weekday='monday')


@pytest.mark.parametrize('max_previous,kwargs,expected', [
    (
        1, dict(microsecond=59867),
        datetime(2001, 1, 1, 0, 0, microsecond=59867),
    ),
    (
        59, dict(second=28, microsecond=0),
        datetime(2023, 1, 1, 1, 59, 28),
    ),
    (
        3600, dict(minute=10, second=20),
        datetime(2016, 6, 1, 12, 10, 20, microsecond=123_456),
    ),
    (
        68400, dict(hour=3),
        datetime(2032, 1, 1, 3, 0, 0, microsecond=123_456),
    ),
    (
        68400 * 60, dict(day=31, minute=59),
        datetime(2032, 3, 31, 0, 59, 0, microsecond=123_456),
    ),
    (
        68400 * 7, dict(weekday='tues', minute=59),
        datetime(2032, 3, 30, 0, 59, 0, microsecond=123_456),
    ),
    (
        68400 * 175,  # previous friday the 13th is February
        dict(day=13, weekday='fri', microsecond=1),
        datetime(2032, 8, 13, 0, 0, 0, microsecond=1),
    ),
    (
        68400 * 365, dict(month=10, day=4, hour=23),
        datetime(2032, 10, 4, 23, 0, 0, microsecond=123_456)),
    (
        1, dict(month=1, day=1, hour=0, minute=0, second=0, microsecond=69875),
        datetime(2001, 1, 1, 0, 0, microsecond=69875),
    ),
])
def test_next_cron_random(max_previous, kwargs, expected):
    for _ in range(100):
        previous = expected - timedelta(seconds=0.9 + random() * max_previous)
        v = next_cron(previous, **kwargs)
        diff = v - previous
        if diff > timedelta(seconds=1):
            assert v == expected


async def foobar():
    return 42


async def test_repr():
    cj = cron(foobar, hour=1, run_at_startup=True)
    assert repr(cj).startswith(
        '<CronJob name=cron:foobar task=<function foobar at',
    )


async def test_str_function():
    cj = cron(foobar, hour=1, run_at_startup=True)
    assert str(cj).startswith('tests.test_cron.foobar')



from typing import Any
from typing import Dict


def default_log_config(verbose: bool) -> Dict[str, Any]:
    """
    Setup default config. for dictConfig.

    :param verbose: level: DEBUG if True, INFO if False
    :return: dict suitable for ``logging.config.dictConfig``
    """
    log_level = 'DEBUG' if verbose else 'INFO'
    return {
        'version': 1,
        'disable_existing_loggers': False,
        'handlers': {
            'darq.standard': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'darq.standard',
            },
        },
        'formatters': {
            'darq.standard': {
                'format': '%(asctime)s: %(message)s',
                'datefmt': '%H:%M:%S',
            },
        },
        'loggers': {
            'darq': {
                'handlers': ['darq.standard'],
                'level': log_level,
            },
        },
    }



import logging

import pytest

from darq import Darq
from darq.app import DarqException
from darq.connections import ArqRedis
from darq.cron import cron
from . import redis_settings
from .test_app import assert_worker_job_finished


async def foobar():
    return 42


async def foobar_custom():
    return 52


async def test_add_cron_jobs(
        darq, caplog, worker_factory, scheduler_factory, arq_redis,
        mocker,
):
    caplog.set_level(logging.INFO)

    with pytest.raises(DarqException):
        darq.add_cron_jobs(cron(foobar))

    mocker.patch('darq.scheduler.to_unix_ms', lambda x: 321)
    foobar_task = darq.task(foobar)
    darq.add_cron_jobs(
        cron('tests.test_scheduler.foobar', run_at_startup=True),
        cron(foobar, run_at_startup=True),
        cron(foobar, run_at_startup=True),  # duplicate doesn't start
        cron(foobar, name='custom_name', run_at_startup=True),
        cron(foobar_task, name='custom_name2', run_at_startup=True),
    )

    with pytest.raises(DarqException):
        # is not CronJob instance
        darq.add_cron_jobs(foobar_task)

    scheduler = scheduler_factory(darq)
    await scheduler.main()
    worker = worker_factory(darq)
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    await worker.main()
    assert worker.jobs_complete == 4
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    function_name = 'tests.test_scheduler.foobar'
    assert_worker_job_finished(
        records=caplog.records,
        job_id='cron:tests.test_scheduler.foobar:321',
        function_name=function_name,
        result='42',
    )
    assert_worker_job_finished(
        records=caplog.records,
        job_id='cron:foobar:321',
        function_name=function_name,
        result='42',
    )
    assert_worker_job_finished(
        records=caplog.records,
        job_id='custom_name:321',
        function_name=function_name,
        result='42',
    )
    assert_worker_job_finished(
        records=caplog.records,
        job_id='custom_name2:321',
        function_name=function_name,
        result='42',
    )


async def test_job_successful_on_specific_queue(
        darq, worker_factory, scheduler_factory, caplog, mocker,
):
    caplog.set_level(logging.INFO)

    foobar_task_default_queue = darq.task(foobar)
    foobar_task_custom_queue = darq.task(foobar_custom, queue='custom_queue')

    mocker.patch('darq.scheduler.to_unix_ms', lambda x: 3215)

    darq.add_cron_jobs(
        cron(foobar_task_default_queue, hour=1, run_at_startup=True),
        cron(foobar_task_custom_queue, hour=1, run_at_startup=True),
    )

    scheduler = scheduler_factory(darq)
    await scheduler.main()

    worker_default_queue = worker_factory(darq)
    await worker_default_queue.main()
    assert worker_default_queue.jobs_complete == 1
    assert worker_default_queue.jobs_failed == 0
    assert worker_default_queue.jobs_retried == 0
    assert_worker_job_finished(
        records=caplog.records,
        job_id='cron:foobar:3215',
        function_name='tests.test_scheduler.foobar',
        result='42',
    )

    worker_custom_queue = worker_factory(darq, queue_name='custom_queue')
    await worker_custom_queue.main()
    assert worker_custom_queue.jobs_complete == 1
    assert worker_custom_queue.jobs_failed == 0
    assert worker_custom_queue.jobs_retried == 0
    assert_worker_job_finished(
        records=caplog.records,
        job_id='cron:foobar_custom:3215',
        function_name='tests.test_scheduler.foobar_custom',
        result='52',
    )


async def test_not_run(darq, worker_factory, caplog, arq_redis):
    caplog.set_level(logging.INFO)

    foobar_task = darq.task(foobar)
    darq.add_cron_jobs(
        cron(foobar_task, hour=1, run_at_startup=False),
    )
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    log = '\n'.join(r.message for r in caplog.records)
    assert 'cron:foobar()' not in log


async def test_startup_shutdown(arq_redis, scheduler_factory):
    calls = []

    async def startup(scheduler_ctx):
        scheduler_ctx['test'] = 123
        assert isinstance(scheduler_ctx.get('redis'), ArqRedis)
        calls.append('startup')

    async def shutdown(scheduler_ctx):
        assert scheduler_ctx['test'] == 123
        assert isinstance(scheduler_ctx.get('redis'), ArqRedis)
        calls.append('shutdown')

    darq = Darq(
        redis_settings=redis_settings, burst=True,
        on_scheduler_startup=startup, on_scheduler_shutdown=shutdown,
    )
    foobar_task = darq.task(foobar)
    darq.add_cron_jobs(
        cron(foobar_task, hour=1),
    )

    scheduler = scheduler_factory(darq)
    await scheduler.main()
    await scheduler.close()

    assert calls == ['startup', 'shutdown']



import asyncio
import dataclasses
import logging
from collections import Counter
from datetime import datetime
from random import shuffle
from time import time

import pytest

from darq.constants import default_queue_name
from darq.jobs import Job
from darq.jobs import JobDef
from darq.jobs import SerializationError
from darq.utils import timestamp_ms
from darq.worker import Retry
from darq.worker import Worker
from .utils import AnyInt
from .utils import CloseToNow


async def foobar():
    return 42


async def test_enqueue_job(darq, arq_redis, worker_factory):
    darq.task(foobar)
    j = await arq_redis.enqueue_job('tests.test_main.foobar', [], {})
    worker: Worker = worker_factory(darq)
    await worker.main()
    r = await j.result(pole_delay=0)
    assert r == 42  # 1


async def test_enqueue_job_different_queues(darq, arq_redis, worker_factory):
    darq.task(foobar)

    j1 = await arq_redis.enqueue_job(
        'tests.test_main.foobar', [], {}, queue_name='arq:queue1',
    )
    j2 = await arq_redis.enqueue_job(
        'tests.test_main.foobar', [], {}, queue_name='arq:queue2',
    )
    worker1 = worker_factory(darq, queue_name='arq:queue1')
    worker2 = worker_factory(darq, queue_name='arq:queue2')

    await worker1.main()
    await worker2.main()
    r1 = await j1.result(pole_delay=0)
    r2 = await j2.result(pole_delay=0)
    assert r1 == 42  # 1
    assert r2 == 42  # 2


async def foobar_error():
    raise RuntimeError('foobar error')


async def test_job_error(darq, arq_redis, worker_factory):
    foobar_error_task = darq.task(foobar_error)
    await darq.connect()

    j = await foobar_error_task.delay()
    worker = worker_factory(darq)
    await worker.main()

    with pytest.raises(RuntimeError, match='foobar error'):
        await j.result(pole_delay=0)


async def test_job_info(arq_redis):
    t_before = time()
    j = await arq_redis.enqueue_job('foobar', [123], {'a': 456})
    info = await j.info()
    assert info.enqueue_time == CloseToNow()
    assert info.job_try is None
    assert info.function == 'foobar'
    assert info.args == [123]
    assert info.kwargs == {'a': 456}
    assert abs(t_before * 1000 - info.score) < 1000


async def test_repeat_job(arq_redis):
    j1 = await arq_redis.enqueue_job('foobar', [], {}, job_id='job_id')
    assert isinstance(j1, Job)
    j2 = await arq_redis.enqueue_job('foobar', [], {}, job_id='job_id')
    assert j2 is None


async def test_defer_until(arq_redis):
    j1 = await arq_redis.enqueue_job(
        'foobar', [], {}, job_id='job_id', defer_until=datetime(2032, 1, 1),
    )
    assert isinstance(j1, Job)
    score = await arq_redis.zscore(default_queue_name, 'job_id')
    assert score == 1_956_528_000_000


async def test_defer_by(arq_redis):
    j1 = await arq_redis.enqueue_job(
        'foobar', [], {}, job_id='job_id', defer_by=20,
    )
    assert isinstance(j1, Job)
    score = await arq_redis.zscore(default_queue_name, 'job_id')
    ts = timestamp_ms()
    assert score > ts + 19000
    assert ts + 21000 > score


async def count(counter, v):
    counter[v] += 1


@pytest.mark.skip(reason='Jobs with "ctx" does not ready')
async def test_mung(darq, arq_redis, worker_factory):
    """
    check a job can't be enqueued multiple times with the same id
    """
    darq.task(count)
    counter = Counter()

    tasks = []
    for i in range(50):
        tasks.extend([
            arq_redis.enqueue_job(
                'tests.test_main.count', args=[counter, i], job_id=f'v-{i}',
            ),
            arq_redis.enqueue_job(
                'tests.test_main.count', args=[counter, i], job_id=f'v-{i}',
            ),
        ])
    shuffle(tasks)
    await asyncio.gather(*tasks)

    worker = worker_factory(darq)
    await worker.main()
    assert counter.most_common(1)[0][1] == 1  # no job go enqueued twice


async def foobar_with_ctx(ctx):
    return ctx['job_try']


@pytest.mark.skip(reason='Jobs with "ctx" does not ready')
async def test_custom_try(darq, arq_redis, worker_factory):
    darq.task(foobar_with_ctx)

    j1 = await arq_redis.enqueue_job('tests.test_main.foobar_with_ctx')
    worker = worker_factory(darq)
    await worker.main()
    r = await j1.result(pole_delay=0)
    assert r == 1

    j2 = await arq_redis.enqueue_job(
        'tests.test_main.foobar_with_ctx', job_try=3,
    )
    await worker.main()
    r = await j2.result(pole_delay=0)
    assert r == 3


async def foobar_ctx_job_try(ctx):
    if ctx['job_try'] == 3:
        raise Retry()
    return ctx['job_try']


@pytest.mark.skip(reason='Jobs with "ctx" does not ready')
async def test_custom_try2(darq, arq_redis, worker_factory):
    darq.task(foobar_ctx_job_try)

    j1 = await arq_redis.enqueue_job(
        'tests.test_main.foobar_ctx_job_try', job_try=3,
    )
    worker = worker_factory(darq)
    await worker.main()
    r = await j1.result(pole_delay=0)
    assert r == 4


class DoesntPickleClass:
    def __getstate__(self):
        raise TypeError("this doesn't pickle")


async def test_cant_pickle_arg(arq_redis, worker_factory):
    class Foobar:
        def __getstate__(self):
            raise TypeError("this doesn't pickle")

    with pytest.raises(
            SerializationError, match='unable to serialize job "foobar"',
    ):
        await arq_redis.enqueue_job('foobar', [DoesntPickleClass()], {})


async def doesnt_pickle():
    return DoesntPickleClass()


async def test_cant_pickle_result(darq, arq_redis, worker_factory):
    darq.task(doesnt_pickle)

    j1 = await arq_redis.enqueue_job('tests.test_main.doesnt_pickle', [], {})
    w = worker_factory(darq)
    await w.main()
    with pytest.raises(SerializationError, match='unable to serialize result'):
        await j1.result(pole_delay=0)


async def test_get_jobs(arq_redis):
    await arq_redis.enqueue_job('foobar', [], dict(a=1, b=2, c=3))
    await asyncio.sleep(0.01)
    await arq_redis.enqueue_job('second', [4], dict(b=5, c=6))
    await asyncio.sleep(0.01)
    await arq_redis.enqueue_job('third', [7], dict(b=8))
    jobs = await arq_redis.queued_jobs()
    assert [dataclasses.asdict(j) for j in jobs] == [
        {
            'function': 'foobar',
            'args': [],
            'kwargs': {'a': 1, 'b': 2, 'c': 3},
            'job_try': None,
            'enqueue_time': CloseToNow(),
            'score': AnyInt(),
        },
        {
            'function': 'second',
            'args': [4],
            'kwargs': {'b': 5, 'c': 6},
            'job_try': None,
            'enqueue_time': CloseToNow(),
            'score': AnyInt(),
        },
        {
            'function': 'third',
            'args': [7],
            'kwargs': {'b': 8},
            'job_try': None,
            'enqueue_time': CloseToNow(),
            'score': AnyInt(),
        },
    ]
    assert jobs[0].score < jobs[1].score < jobs[2].score
    assert isinstance(jobs[0], JobDef)
    assert isinstance(jobs[1], JobDef)
    assert isinstance(jobs[2], JobDef)


async def test_enqueue_multiple(arq_redis, caplog):
    caplog.set_level(logging.DEBUG)
    results = await asyncio.gather(*[
        arq_redis.enqueue_job('foobar', [i], {}, job_id='testing')
        for i in range(10)
    ])
    assert sum(r is not None for r in results) == 1
    assert sum(r is None for r in results) == 9
    assert 'WatchVariableError' not in caplog.text



import asyncio
import logging
import typing as t
from datetime import datetime

from .connections import create_pool
from .connections import RedisSettings
from .types import AnyDict
from .utils import poll
from .utils import to_seconds_strict
from .utils import to_unix_ms

if t.TYPE_CHECKING:  # pragma: no cover
    from darq import Darq
    from .connections import ArqRedis

log = logging.getLogger('darq.scheduler')


class Scheduler:
    """
    :param app: instance of :func:`darq.app.Darq`
    """

    def __init__(self, app: 'Darq') -> None:
        self.cron_jobs = app.cron_jobs
        assert self.cron_jobs and len(self.cron_jobs) > 0, \
            'at least one cron_job must be registered'

        settings = app.worker_settings
        self.app = app
        self.burst = settings.burst
        self.on_startup = app.on_scheduler_startup
        self.on_shutdown = app.on_scheduler_shutdown
        self.poll_delay_s = to_seconds_strict(settings.poll_delay)

        self.pool: 'ArqRedis' = None  # type: ignore[assignment]
        self.redis_settings = app.redis_settings or RedisSettings()

        self.main_task: t.Optional[asyncio.Task[None]] = None
        self.loop = asyncio.get_event_loop()
        self.scheduler_ctx: AnyDict = {}

    def run(self) -> None:
        """
        Sync function to run the scheduler,
        finally closes scheduler connections.
        """
        self.main_task = self.loop.create_task(self.main())
        try:
            self.loop.run_until_complete(self.main_task)
        except asyncio.CancelledError:  # pragma: no cover
            # happens on shutdown, fine
            pass
        finally:
            self.loop.run_until_complete(self.close())

    async def async_run(self) -> None:
        """
        Asynchronously run the scheduler, does not close connections.
        Useful when testing.
        """
        self.main_task = self.loop.create_task(self.main())
        await self.main_task

    async def main(self) -> None:
        log.info(
            'Starting cron scheduler for %d cron jobs: \n%s',
            len(self.cron_jobs), '\n'.join(str(cj) for cj in self.cron_jobs),
        )
        self.pool = await create_pool(self.redis_settings)
        self.scheduler_ctx['redis'] = self.pool
        await self.app.connect(self.pool)
        if self.on_startup:
            await self.on_startup(self.scheduler_ctx)

        async for _ in poll(self.poll_delay_s):
            await self.run_cron()
            if self.burst:
                return

    async def run_cron(self) -> None:
        n = datetime.now()
        job_futures = set()

        for cron_job in self.cron_jobs:
            if cron_job.next_run is None:
                if cron_job.run_at_startup:
                    cron_job.next_run = n
                else:
                    cron_job.set_next(n)

            cron_job.next_run = t.cast(datetime, cron_job.next_run)

            if n >= cron_job.next_run:
                job_id = (
                    f'{cron_job.name}:{to_unix_ms(cron_job.next_run)}'
                    if cron_job.unique
                    else None
                )
                log.info(
                    'Scheduler: Sending due task %s (%s)',
                    cron_job.name, job_id,
                )
                job_futures.add(cron_job.task.apply_async(
                    job_id=job_id,
                ))
                cron_job.set_next(n)

        job_futures and await asyncio.gather(*job_futures)

    async def close(self) -> None:
        if not self.pool:
            return
        if self.on_shutdown:
            await self.on_shutdown(self.scheduler_ctx)
        await self.app.disconnect()
        self.pool.close()
        await self.pool.wait_closed()
        self.pool = None  # type: ignore[assignment]

    def __repr__(self) -> str:
        return f'<Scheduler cron_jobs={len(self.cron_jobs)}>'


def create_scheduler(darq: 'Darq') -> Scheduler:
    return Scheduler(darq)


def run_scheduler(darq: 'Darq') -> Scheduler:
    scheduler = create_scheduler(darq)
    scheduler.run()
    return scheduler



import re
from unittest.mock import patch

from aiohttp.test_utils import loop_context
from click.testing import CliRunner

from darq import Darq
from darq.cli import cli
from darq.cron import cron
from . import redis_settings

darq = Darq(redis_settings=redis_settings, burst=True)


@darq.task
async def foobar(ctx):
    return 42


def test_help():
    runner = CliRunner()
    result = runner.invoke(cli, ['--help'])
    assert result.exit_code == 0
    assert result.output.startswith('Usage: darq [OPTIONS] COMMAND [ARGS]...\n')


def test_worker_run():
    runner = CliRunner()
    with loop_context():
        result = runner.invoke(cli, ['-A', 'tests.test_cli.darq', 'worker'])
    assert result.exit_code == 0
    cli_output = 'Starting worker for 1 functions: tests.test_cli.foobar'
    assert cli_output in result.output


def test_app_bad_argument():
    runner = CliRunner()
    with loop_context():
        result = runner.invoke(cli, ['-A', 'tests.test_cli.foobar', 'worker'])
    assert result.exit_code == 2
    cli_output_re = (
        r'Error: "APP" argument error\. <function foobar at .*> is not '
        r'instance of <class \'darq\.app\.Darq\'>'
    )
    assert re.search(cli_output_re, result.output)


def test_worker_check():
    runner = CliRunner()
    with loop_context():
        result = runner.invoke(cli, [
            '-A', 'tests.test_cli.darq', 'worker', '--check',
        ])
    assert result.exit_code == 1
    expected = 'Health check failed: no health check sentinel value found'
    assert expected in result.output


@patch('darq.worker.Worker')
def test_worker_queue(worker_mock):
    runner = CliRunner()
    custom_queue_name = 'my_queue'
    with loop_context():
        result = runner.invoke(cli, [
            '-A', 'tests.test_cli.darq', 'worker', '--queue', custom_queue_name,
        ])
    assert result.exit_code == 0
    worker_mock.assert_called_once_with(darq, queue_name=custom_queue_name)


@patch('darq.worker.Worker')
def test_worker_max_jobs(worker_mock):
    runner = CliRunner()
    max_jobs = 7
    with loop_context():
        result = runner.invoke(cli, [
            '-A', 'tests.test_cli.darq', 'worker', '--max-jobs', max_jobs,
        ])
    assert result.exit_code == 0
    worker_mock.assert_called_once_with(darq, max_jobs=max_jobs)


async def mock_awatch():
    yield [1]


def test_worker_run_watch(mocker):
    darq.redis_pool = None
    mocker.patch('watchgod.awatch', return_value=mock_awatch())
    runner = CliRunner()
    with loop_context():
        result = runner.invoke(cli, [
            '-A', 'tests.test_cli.darq', 'worker', '--watch', 'tests',
        ])
    assert result.exit_code == 0
    assert '1 files changed, reloading darq worker...' in result.output


async def mock_poll():
    yield


def test_scheduler_run(mocker):
    mocker.patch('darq.scheduler.poll', return_value=mock_awatch())
    runner = CliRunner()

    darq.add_cron_jobs(cron(foobar, hour=1))

    with loop_context():
        result = runner.invoke(cli, ['-A', 'tests.test_cli.darq', 'scheduler'])
    assert result.exit_code == 0
    cli_output = (
        'Starting cron scheduler for 1 cron jobs: \ntests.test_cli.foobar\n'
    )
    assert cli_output in result.output



import asyncio
import logging.config
import os
import sys
import typing as t
from signal import Signals

import click

from .app import Darq
from .logs import default_log_config
from .scheduler import run_scheduler
from .utils import import_string
from .version import __version__
from .worker import check_health
from .worker import create_worker
from .worker import run_worker

health_check_help = 'Health Check: run a health check and exit.'
watch_help = 'Watch a directory and reload the worker upon changes.'
verbose_help = 'Enable verbose output.'


class ContextObject(t.NamedTuple):
    darq: Darq


@click.group('darq')
@click.pass_context
@click.version_option(__version__, '-V', '--version', prog_name='darq')
@click.option('-A', '--app', type=str, required=True)
@click.option('-v', '--verbose', is_flag=True, help=verbose_help)
def cli(ctx: click.Context, *, app: str, verbose: bool) -> None:
    """
    Job queues in python with Asyncio and Redis.

    CLI to run the Darq worker.

    app - path to Darq app instance.
    For example: someproject.darq.darq_app
    """
    sys.path.append(os.getcwd())
    darq = import_string(app)
    if not isinstance(darq, Darq):
        raise click.BadArgumentUsage(
            f'"APP" argument error. {darq!r} is not instance of {Darq!r}',
        )

    ctx.obj = ContextObject(darq)
    logging.config.dictConfig(default_log_config(verbose))


@cli.command()
@click.pass_obj
@click.option('-Q', '--queue', type=str, default=None)
@click.option('--max-jobs', type=int, default=None)
@click.option(
    '--watch', type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help=watch_help,
)
@click.option('--check', is_flag=True, help=health_check_help)
def worker(
        ctx_obj: ContextObject,
        *,
        queue: str,
        max_jobs: int,
        watch: str,
        check: bool,
) -> None:
    """
    CLI to run the Darq worker.
    """
    darq = ctx_obj.darq

    overwrite_settings: t.Dict[str, t.Any] = {}
    if queue is not None:
        overwrite_settings['queue_name'] = queue
    if max_jobs is not None:
        overwrite_settings['max_jobs'] = max_jobs

    if check:
        exit(check_health(darq, queue))
    else:
        if watch:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(
                watch_reload(watch, darq, loop, **overwrite_settings),
            )
        else:
            run_worker(darq, **overwrite_settings)


@cli.command()
@click.pass_obj
def scheduler(ctx_obj: ContextObject) -> None:
    """
    CLI to run the scheduler (cron jobs)
    """
    darq = ctx_obj.darq
    run_scheduler(darq)


async def watch_reload(
        path: str, darq: Darq, loop: asyncio.AbstractEventLoop,
        **overwrite_settings: t.Dict[str, t.Any],
) -> None:
    try:
        from watchgod import awatch
    except ImportError as e:  # pragma: no cover
        raise ImportError(
            'watchgod not installed, use `pip install watchgod`',
        ) from e

    stop_event = asyncio.Event()
    worker = create_worker(darq, **overwrite_settings)

    tasks: t.List[asyncio.Task[t.Any]] = []

    def cancel_tasks() -> None:
        while tasks:
            tasks.pop().cancel()

    def on_stop(sig: Signals) -> None:
        if sig != Signals.SIGUSR1:
            stop_event.set()

    try:
        worker.on_stop = on_stop
        tasks.append(loop.create_task(worker.async_run()))
        async for changes in awatch(path, stop_event=stop_event):
            click.echo(
                f'{len(changes)} files changed, reloading darq worker...',
            )
            worker.handle_sig(Signals.SIGUSR1)
            await worker.close()
            cancel_tasks()
            tasks.append(loop.create_task(worker.async_run()))
        cancel_tasks()
    finally:
        await worker.close()


if __name__ == '__main__':  # pragma: no cover
    cli()



import asyncio
import datetime
import logging
import re
from unittest.mock import call
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from darq import Darq
from darq.app import DarqConnectionError
from darq.connections import ArqRedis
from darq.worker import Task
from . import redis_settings


def assert_is_ctx(ctx):
    assert isinstance(ctx, dict)
    assert isinstance(ctx.get('redis'), ArqRedis)
    assert isinstance(ctx.get('job_id'), str)
    assert isinstance(ctx.get('metadata'), dict)


def parse_log(records):
    return re.sub(
        r'\d+.\d\ds', 'X.XXs',
        '\n'.join(r.message.strip() for r in records),
    )


def assert_worker_job_finished(
        *, records, job_id, function_name, result, args=None, kwargs=None,
):
    call_args = []
    if args:
        call_args.append(', '.join(map(str, args)))
    if kwargs:
        call_args.append(', '.join([
            f'{name}={param}' for name, param in kwargs.items()
        ]))
    call_str = ', '.join(call_args)
    job_id_str = f'{job_id}:' if job_id else ''
    assert (
        f'X.XXs → {job_id_str}{function_name}({call_str})\n'
        f'X.XXs ← {job_id_str}{function_name} ● {result}'
    ) in parse_log(records)


async def foobar(a: int, enqueue_self=False) -> int:
    if enqueue_self:
        await foobar.apply_async([a + 1], {}, job_id='enqueue_self')
    return 42 + a


async def cron_func():
    return 'ok'


async def test_darq_connect_disconnect(darq):
    assert darq.redis_pool is None
    await darq.disconnect()  # should not fail

    await darq.connect()
    assert isinstance(darq.redis_pool, ArqRedis)

    await darq.disconnect()
    assert darq.redis_pool is None


async def test_darq_not_connected(darq):
    foobar_task = darq.task(foobar)
    with pytest.raises(DarqConnectionError):
        await foobar_task.delay()


async def test_job_works_like_a_function(darq):
    foobar_task = darq.task(foobar)
    assert await foobar_task(2) == 44
    assert await foobar_task(a=5) == 47


async def test_task_decorator(darq, caplog, worker_factory):
    caplog.set_level(logging.INFO)

    foobar_task = darq.task(foobar)

    await darq.connect()

    await foobar_task.apply_async([], {'a': 1}, job_id='testing')

    worker = worker_factory(darq)
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    await worker.main()
    assert worker.jobs_complete == 1
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    assert (
        'X.XXs → testing:tests.test_app.foobar(a=1)\n'
        'X.XXs ← testing:tests.test_app.foobar ● 43'
    ) in parse_log(caplog.records)


async def test_task_parametrized(darq):
    assert len(darq.registry) == 0

    timeout = 4
    keep_result = 0.5
    max_tries = 92
    queue = 'my_queue'
    with_ctx = True
    foobar_task = darq.task(
        keep_result=keep_result, timeout=timeout,
        max_tries=max_tries, queue=queue, with_ctx=with_ctx,
    )(foobar)

    task_name = 'tests.test_app.foobar'
    assert len(darq.registry) == 1
    task = darq.registry.get(task_name)
    assert isinstance(task, Task)
    assert task.name == task_name
    assert task.coroutine == foobar_task
    assert task.timeout_s == timeout
    assert task.keep_result_s == keep_result
    assert task.with_ctx == with_ctx


async def test_task_self_enqueue(darq, caplog, worker_factory):
    caplog.set_level(logging.INFO)

    foobar_task = darq.task(foobar)

    await darq.connect()
    await foobar_task.apply_async([1], {'enqueue_self': True}, job_id='testing')

    worker = worker_factory(darq)
    await worker.main()

    assert worker.jobs_complete == 2
    assert (
        'X.XXs → testing:tests.test_app.foobar(1, enqueue_self=True)\n'
        'X.XXs ← testing:tests.test_app.foobar ● 43\n'
        'X.XXs → enqueue_self:tests.test_app.foobar(2)\n'
        'X.XXs ← enqueue_self:tests.test_app.foobar ● 44'
    ) in parse_log(caplog.records)


@pytest.mark.parametrize('func_args,func_kwargs,result', [
    ((1,), {}, 43),
    ((), {'a': 2}, 44),
])
async def test_on_job_callbacks(
        func_args, func_kwargs, result, caplog, worker_factory, arq_redis,
):
    caplog.set_level(logging.INFO)

    expected_metadata = {'test_var': 'ok'}

    async def prepublish_side_effect(
            metadata, task, args, kwargs, job_options,
    ):
        metadata.update(expected_metadata)
        assert isinstance(task, Task)
        assert args == list(func_args)
        assert kwargs == func_kwargs
        assert job_options == {
            'job_id': 'testing',
            'queue_name': None,
            'defer_until': None,
            'defer_by': None,
            'expires': datetime.timedelta(days=1),
            'job_try': None,
        }

    on_job_prerun_future = asyncio.Future()
    on_job_prerun_future.set_result(None)
    on_job_prerun = Mock(return_value=on_job_prerun_future)

    on_job_postrun_future = asyncio.Future()
    on_job_postrun_future.set_result(None)
    on_job_postrun = Mock(return_value=on_job_postrun_future)

    on_job_prepublish = Mock(side_effect=prepublish_side_effect)

    darq = Darq(
        redis_settings=redis_settings,
        burst=True,
        on_job_prerun=on_job_prerun,
        on_job_postrun=on_job_postrun,
        on_job_prepublish=on_job_prepublish,
    )

    foobar_task = darq.task(foobar)

    on_job_prepublish.assert_not_called()

    await darq.connect()
    job_id = 'testing'
    function_name = 'tests.test_app.foobar'
    await foobar_task.apply_async(func_args, func_kwargs, job_id=job_id)

    on_job_prepublish.assert_called_once()

    worker = worker_factory(darq)
    await worker.main()

    assert_worker_job_finished(
        records=caplog.records,
        job_id=job_id,
        function_name=function_name,
        result=result,
        args=func_args,
        kwargs=func_kwargs,
    )

    on_job_prerun.assert_called_once()
    call_args = on_job_prerun.call_args[0]
    assert len(call_args) == 4  # ctx, task, args, kwargs
    ctx = call_args[0]
    assert_is_ctx(ctx)
    assert ctx['metadata'] == expected_metadata
    assert isinstance(call_args[1], Task)
    assert call_args[1].name == function_name
    assert call_args[1].coroutine == foobar_task
    assert call_args[2] == func_args
    assert call_args[3] == func_kwargs

    on_job_postrun.assert_called_once()
    call_args = on_job_postrun.call_args[0]
    assert len(call_args) == 5  # ctx, task, args, kwargs, result
    ctx = call_args[0]
    assert_is_ctx(ctx)
    assert ctx['metadata'] == expected_metadata
    assert isinstance(call_args[1], Task)
    assert call_args[1].name == function_name
    assert call_args[1].coroutine == foobar_task
    assert call_args[2] == func_args
    assert call_args[3] == func_kwargs
    assert call_args[4] == result


@pytest.mark.parametrize(
    'darq_kwargs, task_kwargs, func_args, func_kwargs, apply_async_kwargs,'
    'expected_kwargs', [
        (
            {}, {}, (), {}, {},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': datetime.timedelta(days=1),
            },
        ),
        (
            {'default_job_expires': 3600}, {}, (), {}, {},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': 3600,
            },
        ),
        (
            {'default_job_expires': 3600}, {}, (), {}, {'expires': 32},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': 32,
            },
        ),
        (
            {}, {'expires': 12}, (), {}, {},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': 12,
            },
        ),
        (
            {}, {'expires': 12}, (), {}, {'expires': 200},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': 200,
            },
        ),
    ],
)
@patch('darq.connections.ArqRedis.enqueue_job')
async def test_enqueue_job_params(
        enqueue_job_patched,
        darq_kwargs, task_kwargs, func_args, func_kwargs, apply_async_kwargs,
        expected_kwargs,
        arq_redis,
):
    enqueue_job_patched.reset_mock()

    darq = Darq(redis_settings=redis_settings, **darq_kwargs)
    foobar_task = darq.task(foobar, **task_kwargs)
    await darq.connect()
    await foobar_task.apply_async(func_args, func_kwargs, **apply_async_kwargs)

    enqueue_job_patched.assert_called_once_with(
        'tests.test_app.foobar', func_args, func_kwargs,
        **expected_kwargs,
    )
    await darq.disconnect()


@pytest.mark.parametrize('task_kwargs,apply_async_queue,expected', [
    ({}, None, 'arq:queue'),
    ({'queue': 'my_q'}, None, 'my_q'),
    ({}, 'my_queue', 'my_queue'),
    ({'queue': 'my_q'}, 'new_q', 'new_q'),
])
async def test_task_queue(
        task_kwargs, apply_async_queue, expected,
        arq_redis, darq,
):
    foobar_task = darq.task(foobar, **task_kwargs)
    await darq.connect()

    job = await foobar_task.apply_async([], {}, queue=apply_async_queue)
    assert job._queue_name == expected
    await darq.disconnect()


@pytest.mark.parametrize('packages', [
    (['some.package']),
    (['some.package', 'another.package']),
])
@patch('darq.app.importlib.import_module')
def test_autodiscover_tasks(import_module_mock, darq, packages):
    darq.autodiscover_tasks(packages)
    assert import_module_mock.call_count == len(packages)
    calls = [call(p) for p in packages]
    import_module_mock.assert_has_calls(calls)


@pytest.mark.parametrize(
    'darq_kwargs, task_kwargs, apply_async_kwargs,'
    'expected_kwargs', [
        (
            {}, {}, {},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': datetime.timedelta(days=1),
            },
        ),
        (
            {'default_job_expires': None}, {}, {},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': None,
            },
        ),
        (
            {}, {'expires': None}, {},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': None,
            },
        ),
        (
            {}, {'expires': 111}, {'expires': 222},
            {
                'defer_by': None, 'defer_until': None, 'job_id': None,
                'job_try': None, 'queue_name': None,
                'expires': 222,
            },
        ),
    ],
)
@patch('darq.connections.ArqRedis.enqueue_job')
async def test_expires_param(
        enqueue_job_patched,
        darq_kwargs, task_kwargs, apply_async_kwargs,
        expected_kwargs,
        arq_redis,
):
    enqueue_job_patched.reset_mock()

    darq = Darq(redis_settings=redis_settings, **darq_kwargs)
    foobar_task = darq.task(foobar, **task_kwargs)
    await darq.connect()
    await foobar_task.apply_async(tuple(), {}, **apply_async_kwargs)

    enqueue_job_patched.assert_called_once_with(
        'tests.test_app.foobar', tuple(), {},
        **expected_kwargs,
    )
    await darq.disconnect()


async def foobar_with_ctx(ctx, a: int) -> int:
    return 42 + a + ctx['b']


@pytest.mark.parametrize('func_args,func_kwargs,func_ctx,result', [
    ((1,), {}, {'b': 1}, 44),
    ((), {'a': 2}, {'b': 1}, 45),
])
async def test_run_task_with_ctx(
        func_args, func_kwargs, func_ctx, result,
        arq_redis, caplog, worker_factory,
):
    caplog.set_level(logging.INFO)

    async def on_worker_startup(ctx):
        ctx.update(func_ctx)

    darq = Darq(
        redis_settings=redis_settings,
        burst=True,
        on_startup=on_worker_startup,
    )

    foobar_with_ctx_task = darq.task(foobar_with_ctx, with_ctx=True)

    await darq.connect()

    job_id = 'testing'
    function_name = 'tests.test_app.foobar_with_ctx'
    await foobar_with_ctx_task.apply_async(
        func_args, func_kwargs, job_id=job_id,
    )

    worker = worker_factory(darq)
    await worker.main()

    assert_worker_job_finished(
        records=caplog.records,
        job_id=job_id,
        function_name=function_name,
        result=result,
        args=func_args,
        kwargs=func_kwargs,
    )



default_queue_name = 'arq:queue'
job_key_prefix = 'arq:job:'
in_progress_key_prefix = 'arq:in-progress:'
result_key_prefix = 'arq:result:'
retry_key_prefix = 'arq:retry:'
health_check_key_suffix = ':health-check'



import asyncio
import functools
import logging
import re
import signal
from unittest.mock import MagicMock

import msgpack
import pytest
from aiohttp.test_utils import loop_context
from aioredis import create_redis_pool

from darq import Darq
from darq.constants import default_queue_name
from darq.constants import health_check_key_suffix
from darq.constants import job_key_prefix
from darq.jobs import Job
from darq.jobs import JobStatus
from darq.worker import async_check_health
from darq.worker import check_health
from darq.worker import FailedJobs
from darq.worker import JobExecutionFailed
from darq.worker import Retry
from darq.worker import run_worker
from . import redis_settings


async def foobar():
    return 42


async def foo(v):
    return v + 1


async def bar(v):
    await foo.delay(v + 1)


async def raise_on_value(v=0):
    if v > 0:
        raise ValueError('xxx')


async def return_error():
    return TypeError('xxx')


async def retry(defer=None):
    raise Retry(defer=defer)


async def fails():
    raise TypeError('my type error')


@pytest.mark.skip
async def test_handle_sig_delayed(
        darq, caplog, arq_redis, worker_factory, loop,
):
    darq.task(foobar)
    caplog.set_level(logging.INFO)

    worker = worker_factory(darq)

    long_running_task_mock = MagicMock(done=MagicMock(return_value=False))
    worker.main_task = MagicMock()
    worker.tasks = [
        MagicMock(done=MagicMock(return_value=True)),
        long_running_task_mock,
    ]

    if caplog.records:
        assert all(
            'Task was destroyed but it is pending!' in record.message
            for record in caplog.records
        )
    worker.handle_sig(signal.SIGINT)
    await asyncio.sleep(0)
    assert caplog.records[-1].message == (
        'Warm shutdown. Awaiting for 1 jobs with 30 seconds timeout.'
    )
    if len(caplog.records) > 1:
        assert all(
            'Task was destroyed but it is pending!' in record.message
            for record in caplog.records[:-1]
        )
    assert worker.main_task.cancel.call_count == 0
    assert worker.tasks[0].cancel.call_count == 0
    assert worker.tasks[1].cancel.call_count == 0
    await asyncio.sleep(0)

    worker.tasks[1].done.return_value = True
    await asyncio.sleep(1)

    assert 'shutdown on SIGINT' in caplog.records[-1].message
    assert worker.main_task.cancel.call_count == 1
    assert worker.tasks[0].cancel.call_count == 0
    assert worker.tasks[1].cancel.call_count == 0


async def test_handle_sig_hard(darq, caplog, worker_factory):
    darq.task(foobar)
    caplog.set_level(logging.INFO)

    worker = worker_factory(darq)

    long_running_task_mock = MagicMock(done=MagicMock(return_value=False))
    worker.main_task = MagicMock()
    worker.tasks = [long_running_task_mock]

    caplog.clear()
    worker.handle_sig(signal.SIGINT)
    await asyncio.sleep(0)
    assert len(caplog.records) == 1
    assert caplog.records[0].message == (
        'Warm shutdown. Awaiting for 1 jobs with 30 seconds timeout.'
    )
    worker.handle_sig(signal.SIGINT)
    assert caplog.records[1].message == (
        'shutdown on SIGINT ◆ 0 jobs complete ◆ 0 failed ◆ 0 retries ◆ '
        '1 ongoing to cancel'
    )


def test_no_jobs(darq, arq_redis, event_loop):
    darq.task(foobar)

    event_loop.run_until_complete(
        arq_redis.enqueue_job('tests.test_worker.foobar', [], {}),
    )
    with loop_context():
        worker = run_worker(darq)
    assert worker.jobs_complete == 1
    assert str(worker) == (
        '<Worker j_complete=1 j_failed=0 j_retried=0 j_ongoing=0>'
    )


def test_health_check_direct(darq):
    with loop_context():
        assert check_health(darq, queue=None) == 1


async def test_health_check_fails():
    assert 1 == await async_check_health(redis_settings)


async def test_health_check_pass(arq_redis):
    await arq_redis.set(default_queue_name + health_check_key_suffix, b'1')
    assert 0 == await async_check_health(redis_settings)


async def test_set_health_check_key(arq_redis, worker_factory):
    darq = Darq(
        redis_settings=redis_settings, burst=True, poll_delay=0,
        health_check_key='darq:test:health-check',
    )
    darq.task(foobar)
    await arq_redis.enqueue_job(
        'tests.test_worker.foobar', [], {}, job_id='testing',
    )
    worker = worker_factory(darq)
    await worker.main()
    assert sorted(await arq_redis.keys('*')) == [
        'arq:result:testing',
        'darq:test:health-check',
    ]


async def test_job_successful(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    foobar_task = darq.task(foobar)
    await darq.connect()

    await foobar_task.apply_async([], {}, job_id='testing')
    worker = worker_factory(darq)
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    await worker.main()
    assert worker.jobs_complete == 1
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    log = re.sub(
        r'\d+.\d\ds', 'X.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    expected = (
        'X.XXs → testing:tests.test_worker.foobar()\n  '
        'X.XXs ← testing:tests.test_worker.foobar ● 42'
    )
    assert expected in log


@pytest.mark.skip('Jobs with "ctx" does not ready')
async def test_job_retry(darq, arq_redis, worker_factory, caplog):
    async def retry(ctx):
        if ctx['job_try'] <= 2:
            raise Retry(defer=0.01)

    caplog.set_level(logging.INFO)
    await arq_redis.enqueue_job('retry', _job_id='testing')
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 1
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 2

    log = re.sub(
        r'(\d+).\d\ds', r'\1.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    assert '0.XXs ↻ testing:retry retrying job in 0.XXs\n' in log
    assert '0.XXs → testing:retry() try=2\n' in log
    assert '0.XXs ← testing:retry ●' in log


async def test_job_retry_dont_retry(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    retry_task = darq.task(retry)
    await darq.connect()

    await retry_task.apply_async([], {'defer': 0.01}, job_id='testing')
    worker = worker_factory(darq)
    with pytest.raises(FailedJobs) as exc_info:
        await worker.run_check(retry_jobs=False)
    assert str(exc_info.value) == '1 job failed <Retry defer 0.01s>'

    assert '↻' not in caplog.text
    expected = (
        '! testing:tests.test_worker.retry failed, Retry: <Retry defer 0.01s>\n'
    )
    assert expected in caplog.text


async def test_job_retry_max_jobs(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    retry_task = darq.task(retry)
    await darq.connect()

    await retry_task.apply_async([], {'defer': 0.01}, job_id='testing')
    worker = worker_factory(darq)
    assert await worker.run_check(max_burst_jobs=1) == 0
    assert worker.jobs_complete == 0
    assert worker.jobs_retried == 1
    assert worker.jobs_failed == 0

    log = re.sub(r'(\d+).\d\ds', r'\1.XXs', caplog.text)
    assert (
        '0.XXs ↻ testing:tests.test_worker.retry retrying job in 0.XXs\n' in log
    )
    assert '0.XXs → testing:tests.test_worker.retry() try=2\n' not in log


async def test_job_job_not_found(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    darq.task(foobar)
    await darq.connect()

    await arq_redis.enqueue_job('missing', [], {}, job_id='testing')
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 1
    assert worker.jobs_retried == 0

    log = re.sub(
        r'\d+.\d\ds', 'X.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    assert "job testing, function 'missing' not found" in log


async def test_job_job_not_found_run_check(
        darq, arq_redis, worker_factory, caplog,
):
    darq.task(foobar)
    caplog.set_level(logging.INFO)
    await arq_redis.enqueue_job('missing', [], {}, job_id='testing')
    worker = worker_factory(darq)
    with pytest.raises(FailedJobs) as exc_info:
        await worker.run_check()

    assert exc_info.value.count == 1
    assert len(exc_info.value.job_results) == 1
    failure = exc_info.value.job_results[0].result
    assert failure == JobExecutionFailed("function 'missing' not found")
    assert failure != 123  # check the __eq__ method of JobExecutionFailed


async def test_retry_lots(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    retry_task = darq.task(retry)
    await darq.connect()

    await retry_task.apply_async([], {}, job_id='testing')
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 1
    assert worker.jobs_retried == 5

    log = re.sub(
        r'\d+.\d\ds', 'X.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    expected = (
        '  X.XXs ! testing:tests.test_worker.retry max retries 5 exceeded'
    )
    assert expected in log


async def test_retry_lots_without_keep_result(darq, arq_redis, worker_factory):
    retry_task = darq.task(retry, keep_result=0)
    await darq.connect()
    await retry_task.apply_async([], {}, job_id='testing')
    worker = worker_factory(darq)
    await worker.main()  # Should not raise MultiExecError


async def test_retry_lots_check(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    retry_task = darq.task(retry)
    await darq.connect()

    await retry_task.apply_async([], {}, job_id='testing')
    worker = worker_factory(darq)
    with pytest.raises(FailedJobs, match='max 5 retries exceeded'):
        await worker.run_check()


@pytest.mark.skip(reason='Jobs with "ctx" does not ready')
async def test_cancel_error(darq, arq_redis, worker_factory, caplog):
    async def retry(ctx):
        if ctx['job_try'] == 1:
            raise asyncio.CancelledError()

    caplog.set_level(logging.INFO)
    await arq_redis.enqueue_job('retry', _job_id='testing')
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 1
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 1

    log = re.sub(
        r'\d+.\d\ds', 'X.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    assert 'X.XXs ↻ testing:retry cancelled, will be run again' in log


async def test_job_expired(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    foobar_task = darq.task(foobar)
    await darq.connect()

    await foobar_task.apply_async([], {}, job_id='testing')
    await arq_redis.delete(job_key_prefix + 'testing')
    worker = worker_factory(darq)
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 1
    assert worker.jobs_retried == 0

    log = re.sub(
        r'\d+.\d\ds', 'X.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    assert 'job testing expired' in log


async def test_job_expired_run_check(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    darq_task = darq.task(foobar)
    await darq.connect()

    await darq_task.apply_async([], {}, job_id='testing')
    await arq_redis.delete(job_key_prefix + 'testing')
    worker = worker_factory(darq)
    with pytest.raises(FailedJobs) as exc_info:
        await worker.run_check()

    exc_value = exc_info.value
    assert str(exc_value) == "1 job failed JobExecutionFailed('job expired')"
    assert exc_value.count == 1
    assert len(exc_value.job_results) == 1
    assert exc_value.job_results[0].result == JobExecutionFailed('job expired')


async def test_job_old(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    foobar_task = darq.task(foobar)
    await darq.connect()

    await foobar_task.apply_async([], {}, job_id='testing', defer_by=-2)
    worker = worker_factory(darq)
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    await worker.main()
    assert worker.jobs_complete == 1
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    log = re.sub(
        r'(\d+).\d\ds', r'\1.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    assert log.endswith(
        '  0.XXs → testing:tests.test_worker.foobar() delayed=2.XXs\n'
        '  0.XXs ← testing:tests.test_worker.foobar ● 42',
    )


async def test_retry_repr():
    assert str(Retry(123)) == '<Retry defer 123.00s>'


async def test_str_function(darq, arq_redis, worker_factory, caplog):
    darq.task(asyncio.sleep)
    caplog.set_level(logging.INFO)
    await arq_redis.enqueue_job('asyncio.tasks.sleep', [], {}, job_id='testing')
    worker = worker_factory(darq)
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 1
    assert worker.jobs_retried == 0

    log = re.sub(
        r'(\d+).\d\ds', r'\1.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    assert '0.XXs ! testing:asyncio.tasks.sleep failed, TypeError' in log


async def test_startup_shutdown(arq_redis, worker_factory):
    calls = []

    async def startup(ctx):
        calls.append('startup')

    async def shutdown(ctx):
        calls.append('shutdown')

    darq = Darq(
        redis_settings=redis_settings, burst=True,
        on_startup=startup, on_shutdown=shutdown,
    )
    foobar_task = darq.task(foobar)
    await darq.connect()

    await foobar_task.apply_async([], {}, job_id='testing')
    worker = worker_factory(darq)
    await worker.main()
    await worker.close()

    assert calls == ['startup', 'shutdown']


class CustomError(RuntimeError):
    def extra(self):
        return {'x': 'y'}


async def error_function():
    raise CustomError('this is the error')


async def test_exc_extra(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    error_function_task = darq.task(error_function)
    await darq.connect()

    await error_function_task.apply_async([], {}, job_id='testing')
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_failed == 1

    log = re.sub(
        r'(\d+).\d\ds', r'\1.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    expected = (
        '0.XXs ! testing:tests.test_worker.error_function failed, '
        'CustomError: this is the error'
    )
    assert expected in log
    error = next(r for r in caplog.records if r.levelno == logging.ERROR)
    assert error.extra == {'x': 'y'}


async def unpickleable():
    class Foo:
        pass
    return Foo()


async def test_unpickleable(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    darq.task(unpickleable)

    await arq_redis.enqueue_job(
        'tests.test_worker.unpickleable', [], {}, job_id='testing',
    )
    worker = worker_factory(darq)
    await worker.main()

    log = re.sub(
        r'(\d+).\d\ds', r'\1.XXs',
        '\n'.join(r.message for r in caplog.records),
    )
    expected = (
        'error serializing result of testing:tests.test_worker.unpickleable'
    )
    assert expected in log


async def test_log_health_check(arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    darq = Darq(
        redis_settings=redis_settings, burst=True, health_check_interval=0,
    )
    foobar_task = darq.task(foobar)
    await darq.connect()

    await foobar_task.apply_async([], {}, job_id='testing')
    worker = worker_factory(darq)
    await worker.main()
    await worker.main()
    await worker.main()
    assert worker.jobs_complete == 1

    expected = 'j_complete=1 j_failed=0 j_retried=0 j_ongoing=0 queued=0'
    assert expected in caplog.text
    # can happen more than once due to redis pool size
    # assert log.count('recording health') == 1
    assert 'recording health' in caplog.text


async def test_remain_keys(darq, arq_redis, worker_factory):
    foobar_task = darq.task(foobar)
    await darq.connect()

    redis2 = await create_redis_pool(
        (redis_settings.host, redis_settings.port), encoding='utf8',
    )
    try:
        await foobar_task.apply_async([], {}, job_id='testing')
        assert sorted(await redis2.keys('*')) == [
            'arq:job:testing', 'arq:queue',
        ]
        worker = worker_factory(darq)
        await worker.main()
        assert sorted(await redis2.keys('*')) == [
            'arq:queue:health-check', 'arq:result:testing',
        ]
        await worker.close()
        assert sorted(await redis2.keys('*')) == ['arq:result:testing']
    finally:
        redis2.close()
        await redis2.wait_closed()


async def test_remain_keys_no_results(darq, arq_redis, worker_factory):
    foobar_task = darq.task(foobar, keep_result=0)
    await darq.connect()

    await foobar_task.apply_async([], {}, job_id='testing')
    assert sorted(await arq_redis.keys('*')) == ['arq:job:testing', 'arq:queue']
    worker = worker_factory(darq)
    await worker.main()
    assert sorted(await arq_redis.keys('*')) == ['arq:queue:health-check']


async def test_run_check_passes(darq, arq_redis, worker_factory):
    foobar_task = darq.task(foobar)
    await darq.connect()

    await foobar_task.apply_async([], {})
    await foobar_task.apply_async([], {})
    worker = worker_factory(darq)
    assert 2 == await worker.run_check()


async def test_run_check_error(darq, arq_redis, worker_factory):
    fails_task = darq.task(fails)
    await darq.connect()

    await fails_task.delay()
    worker = worker_factory(darq)
    with pytest.raises(
            FailedJobs, match=r"1 job failed TypeError\('my type error'",
    ):
        await worker.run_check()


async def test_run_check_error2(darq, arq_redis, worker_factory):
    raise_on_value_task = darq.task(raise_on_value)
    await darq.connect()

    await raise_on_value_task.delay(1)
    await raise_on_value_task.delay(1)
    worker = worker_factory(darq)
    with pytest.raises(FailedJobs, match='2 jobs failed:\n') as exc_info:
        await worker.run_check()
    assert len(exc_info.value.job_results) == 2


async def test_return_exception(darq, arq_redis, worker_factory):
    return_error_task = darq.task(return_error)
    await darq.connect()

    j = await return_error_task.delay()
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 1
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    r = await j.result(pole_delay=0)
    assert isinstance(r, TypeError)
    info = await j.result_info()
    assert info.success is True


async def test_error_success(darq, arq_redis, worker_factory):
    fails_task = darq.task(fails)
    await darq.connect()

    j = await fails_task.delay()
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 1
    assert worker.jobs_retried == 0
    info = await j.result_info()
    assert info.success is False


async def test_many_jobs_expire(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    foobar_task = darq.task(foobar)
    await darq.connect()

    await foobar_task.delay()
    await asyncio.gather(*[
        arq_redis.zadd(default_queue_name, 1, f'testing-{i}')
        for i in range(100)
    ])
    worker = worker_factory(darq)
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0
    await worker.main()
    assert worker.jobs_complete == 1
    assert worker.jobs_failed == 100
    assert worker.jobs_retried == 0

    log = '\n'.join(r.message for r in caplog.records)
    assert 'job testing-0 expired' in log
    assert log.count(' expired') == 100


async def test_repeat_job_result(darq, arq_redis, worker_factory):
    foobar_task = darq.task(foobar)
    await darq.connect()

    j1 = await foobar_task.apply_async([], {}, job_id='job_id')
    assert isinstance(j1, Job)
    assert await j1.status() == JobStatus.queued

    j2 = await foobar_task.apply_async([], {}, job_id='job_id')
    assert j2 is None

    await worker_factory(darq).run_check()
    assert await j1.status() == JobStatus.complete

    j3 = await foobar_task.apply_async([], {}, job_id='job_id')
    assert j3 is None


async def test_queue_read_limit_equals_max_jobs(arq_redis, worker_factory):
    darq = Darq(redis_settings=redis_settings, burst=True, max_jobs=2)
    foobar_task = darq.task(foobar)
    await darq.connect()

    for _ in range(4):
        await foobar_task.delay()

    assert await arq_redis.zcard(default_queue_name) == 4
    worker = worker_factory(darq)
    worker.pool = arq_redis
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    await worker._poll_iteration()
    await asyncio.sleep(0.1)
    assert await arq_redis.zcard(default_queue_name) == 2
    assert worker.jobs_complete == 2
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    await worker._poll_iteration()
    await asyncio.sleep(0.1)
    assert await arq_redis.zcard(default_queue_name) == 0
    assert worker.jobs_complete == 4
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0


async def test_custom_queue_read_limit(arq_redis, worker_factory):
    darq = Darq(
        redis_settings=redis_settings, burst=True, max_jobs=4,
        queue_read_limit=2,
    )
    foobar_task = darq.task(foobar)
    await darq.connect()

    for _ in range(4):
        await foobar_task.delay()

    assert await arq_redis.zcard(default_queue_name) == 4
    worker = worker_factory(darq)
    worker.pool = arq_redis
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    await worker._poll_iteration()
    await asyncio.sleep(0.1)
    assert await arq_redis.zcard(default_queue_name) == 2
    assert worker.jobs_complete == 2
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0

    await worker._poll_iteration()
    await asyncio.sleep(0.1)
    assert await arq_redis.zcard(default_queue_name) == 0
    assert worker.jobs_complete == 4
    assert worker.jobs_failed == 0
    assert worker.jobs_retried == 0


async def test_custom_serializers(arq_redis_msgpack, worker_factory):
    darq = Darq(
        redis_settings=redis_settings, burst=True,
        job_serializer=msgpack.packb,
        job_deserializer=functools.partial(msgpack.unpackb, raw=False),
    )
    foobar_task = darq.task(foobar)
    await darq.connect()

    j = await foobar_task.apply_async([], {}, job_id='job_id')
    worker = worker_factory(darq)
    info = await j.info()
    assert info.function == 'tests.test_worker.foobar'
    assert await worker.run_check() == 1
    assert await j.result() == 42
    r = await j.info()
    assert r.result == 42


class UnpickleFails:
    def __init__(self, v):
        self.v = v

    def __setstate__(self, state):
        raise ValueError('this broke')


async def test_deserialization_error(darq, arq_redis, worker_factory):
    darq.task(foobar)
    await arq_redis.enqueue_job(
        'tests.test_worker.foobar',
        [UnpickleFails('hello')], {}, job_id='job_id',
    )
    worker = worker_factory(darq)
    with pytest.raises(FailedJobs) as exc_info:
        await worker.run_check()
    assert str(exc_info.value) == (
        "1 job failed DeserializationError('unable to deserialize job')"
    )


async def test_incompatible_serializers_1(
        darq, arq_redis_msgpack, worker_factory,
):
    darq.task(foobar)
    await arq_redis_msgpack.enqueue_job(
        'tests.test_worker.foobar', [], {}, job_id='job_id',
    )
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 1
    assert worker.jobs_retried == 0


async def test_incompatible_serializers_2(arq_redis, worker_factory):
    darq = Darq(
        redis_settings=redis_settings, burst=True,
        job_serializer=msgpack.packb,
        job_deserializer=functools.partial(msgpack.unpackb, raw=False),
    )
    darq.task(foobar)

    await arq_redis.enqueue_job(
        'tests.test_worker.foobar', [], {}, job_id='job_id',
    )
    worker = worker_factory(darq)
    await worker.main()
    assert worker.jobs_complete == 0
    assert worker.jobs_failed == 1
    assert worker.jobs_retried == 0


async def test_max_jobs_completes(darq, arq_redis, worker_factory):
    raise_on_value_task = darq.task(raise_on_value)
    await darq.connect()

    await raise_on_value_task.delay()
    await raise_on_value_task.delay(1)
    await raise_on_value_task.delay(2)
    worker = worker_factory(darq)
    with pytest.raises(FailedJobs) as exc_info:
        await worker.run_check(max_burst_jobs=3)
    assert repr(exc_info.value).startswith('<2 jobs failed:')


async def test_max_bursts_sub_call(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    darq.task(foo)
    bar_task = darq.task(bar)
    await darq.connect()

    await bar_task.delay(10)
    worker = worker_factory(darq)
    assert await worker.run_check(max_burst_jobs=1) == 1
    assert worker.jobs_complete == 1
    assert worker.jobs_retried == 0
    assert worker.jobs_failed == 0
    assert 'bar(10)' in caplog.text
    assert 'foo' in caplog.text


async def test_max_bursts_multiple(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.INFO)
    foo_task = darq.task(foo)
    await darq.connect()

    await foo_task.delay(1)
    await asyncio.sleep(0.005)
    await foo_task.delay(2)
    worker = worker_factory(darq)
    assert await worker.run_check(max_burst_jobs=1) == 1
    assert worker.jobs_complete == 1
    assert worker.jobs_retried == 0
    assert worker.jobs_failed == 0
    assert 'foo(1)' in caplog.text
    assert 'foo(2)' not in caplog.text


async def test_max_bursts_dont_get(darq, arq_redis, worker_factory):
    foo_task = darq.task(foo)
    await darq.connect()

    await foo_task.delay(1)
    await foo_task.delay(2)
    await darq.connect()
    worker = worker_factory(darq)

    worker.max_burst_jobs = 0
    assert len(worker.tasks) == 0
    await worker._poll_iteration()
    assert len(worker.tasks) == 0


async def test_non_burst(darq, arq_redis, worker_factory, caplog, event_loop):
    caplog.set_level(logging.INFO)
    foo_task = darq.task(foo)
    await darq.connect()

    await foo_task.apply_async([1], {}, job_id='testing')
    worker = worker_factory(darq)
    worker.burst = False
    t = event_loop.create_task(worker.main())
    await asyncio.sleep(0.1)
    t.cancel()
    assert worker.jobs_complete == 1
    assert worker.jobs_retried == 0
    assert worker.jobs_failed == 0
    assert '← testing:tests.test_worker.foo ● 2' in caplog.text


@pytest.mark.skip(reason='Not working')
async def test_multi_exec(darq, arq_redis, worker_factory, caplog):
    caplog.set_level(logging.DEBUG, logger='darq.worker')
    darq.task(foo)
    await arq_redis.enqueue_job(
        'tests.test_worker.foo_multi_exec', 1, _job_id='testing',
    )
    await darq.connect()
    worker = worker_factory(darq)
    await asyncio.gather(*[worker.run_jobs(['testing']) for _ in range(5)])
    assert (
        'multi-exec error, job testing already started elsewhere' in caplog.text
    )
    assert 'WatchVariableError' not in caplog.text



import datetime
import sys
import typing as t

if sys.version_info >= (3, 8):
    from typing import Protocol
    from typing import TypedDict
else:
    from typing_extensions import Protocol
    from typing_extensions import TypedDict

if t.TYPE_CHECKING:  # pragma: no cover
    import darq
    from darq.jobs import Job  # noqa F401
    from darq.worker import Task  # noqa F401

UNSET_ARG = t.NewType('UNSET_ARG', object)
unset_arg: UNSET_ARG = UNSET_ARG(object())

WrappingFunc = t.TypeVar(
    'WrappingFunc', bound=t.Callable[..., t.Coroutine[t.Any, t.Any, t.Any]],
)
T = t.TypeVar('T')


class DarqTask(Protocol[T]):
    __call__: T
    delay: T
    __name__: str
    __qualname__: str

    @staticmethod
    async def apply_async(
            args: t.Optional[t.Sequence[t.Any]] = None,
            kwargs: t.Optional[t.Mapping[str, t.Any]] = None,
            *,
            job_id: t.Optional[str] = None,
            queue: t.Optional[str] = None,
            defer_until: t.Optional[datetime.datetime] = None,
            defer_by: t.Optional['AnyTimedelta'] = None,
            expires: t.Union[None, 'AnyTimedelta', UNSET_ARG] = unset_arg,
            job_try: t.Optional[int] = None,
    ) -> t.Optional['Job']:  # pragma: no cover
        ...


AnyCallable = t.Callable[..., t.Any]
CoroutineType = t.Callable[..., t.Coroutine[t.Any, t.Any, t.Any]]
AnyTimedelta = t.Union[int, float, datetime.timedelta]
AnyDict = t.Dict[t.Any, t.Any]
DataDict = t.Dict[str, t.Any]
ArgsType = t.Sequence[t.Any]
KwargsType = t.Mapping[str, t.Any]
MutableArgsType = t.List[t.Any]
MutableKwargsType = t.Dict[str, t.Any]


class JobCtx(TypedDict):
    redis: 'darq.connections.ArqRedis'
    job_id: str
    job_try: int
    enqueue_time: datetime.datetime
    score: int
    metadata: DataDict


class JobEnqueueOptions(TypedDict):
    job_id: t.Optional[str]
    queue_name: t.Optional[str]
    defer_until: t.Optional[datetime.datetime]
    defer_by: t.Optional[AnyTimedelta]
    expires: t.Optional[AnyTimedelta]
    job_try: t.Optional[int]


OnJobPrerunType = t.Callable[
    [AnyDict, 'Task', ArgsType, KwargsType],
    t.Awaitable[None],
]
OnJobPostrunType = t.Callable[
    [AnyDict, 'Task', ArgsType, KwargsType, t.Any],
    t.Awaitable[None],
]
OnJobPrepublishType = t.Callable[
    [
        DataDict, 'Task', MutableArgsType, MutableKwargsType,
        JobEnqueueOptions,
    ],
    t.Awaitable[None],
]



import pathlib
from importlib.machinery import SourceFileLoader

import setuptools

version = SourceFileLoader('version', 'darq/version.py').load_module()


def readfile(filename: str) -> str:
    return pathlib.Path(filename).read_text('utf-8').strip()


long_description = '\n\n'.join((
    readfile('README.rst'),
    readfile('CHANGES.rst'),
))

setuptools.setup(
    name='darq',
    version=version.__version__,
    author='Igor Mozharovsky',
    author_email='igor.mozharovsky@gmail.com',
    description=(
        'Async task manager with Celery-like features and type checking. '
        'Fork of arq.'),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/seedofjoy/darq',
    packages=['darq'],
    package_data={'darq': ['py.typed']},
    data_files=['CHANGES.rst'],
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Typing :: Typed',
    ],
    entry_points={
        'console_scripts': ['darq = darq.cli:cli'],
    },
    install_requires=[
        'async-timeout>=3.0.0, <5.0',
        'click>=6.7',
        'typing_extensions>=4.0; python_version<"3.8"',
    ],
    extras_require={
        'watch': ['watchgod>=0.4'],
    },
    python_requires='>=3.7',
    license='MIT',
)



import asyncio
import datetime
import importlib
import typing as t

from .connections import ArqRedis
from .connections import create_pool
from .connections import RedisSettings
from .constants import default_queue_name
from .cron import CronJob
from .jobs import Deserializer
from .jobs import Job
from .jobs import Serializer
from .registry import Registry
from .types import AnyTimedelta
from .types import DarqTask
from .types import DataDict
from .types import JobEnqueueOptions
from .types import OnJobPostrunType
from .types import OnJobPrepublishType
from .types import OnJobPrerunType
from .types import UNSET_ARG
from .types import unset_arg
from .types import WrappingFunc
from .utils import get_function_name
from .utils import SecondsTimedelta
from .worker import Task
from .worker import WorkerSettings

CtxType = t.Dict[t.Any, t.Any]
TD_1_DAY = datetime.timedelta(days=1)


class DarqException(Exception):
    pass


class DarqConnectionError(DarqException):
    pass


class DarqConfigError(DarqException):
    pass


class Darq:
    """
    Darq application.

    :param queue_name: queue name to get tasks from
    :param redis_settings: settings for creating a redis connection
    :param burst: whether to stop the worker once all tasks have been run
    :param on_startup: coroutine function to run at worker startup
    :param on_shutdown: coroutine function to run at worker shutdown
    :param on_job_prerun: coroutine function to run before task starts
    :param on_job_postrun: coroutine function to run after task finish
    :param on_job_prepublish: coroutine function to run before enqueue task
    :param max_jobs: maximum number of tasks to run at a time
    :param job_timeout: default task timeout (max run time)
    :param keep_result: default duration to keep task results for
    :param poll_delay: duration between polling the queue for new tasks
    :param queue_read_limit: the maximum number of tasks to pull from the queue
                             each time it's polled;
                             by default it equals ``max_jobs``
    :param max_tries: default maximum number of times to retry a task
    :param health_check_interval: how often to set the health check key
    :param health_check_key: redis key under which health check is set
    :param ctx: dict object, data from it will be pass to hooks:
                ``on_startup``, ``on_shutdown`` - can modify ``ctx``;
                ``on_job_prerun``, ``on_job_postrun`` - readonly
    :param retry_jobs: whether to retry tasks on Retry or CancelledError or not
    :param max_burst_jobs: the maximum number of tasks to process in burst mode
                           (disabled with negative values)
    :param job_serializer: a function that serializes Python objects to bytes,
                           defaults to pickle.dumps
    :param job_deserializer: a function that deserializes bytes into Python
                             objects, defaults to pickle.loads
    :param default_job_expires: default task expires. If the task still hasn't
                                started after this duration, do not run it
    """

    def __init__(
            self,
            *,
            queue_name: str = default_queue_name,
            redis_settings: t.Optional[RedisSettings] = None,
            burst: bool = False,
            on_startup: t.Callable[[CtxType], t.Awaitable[None]] = None,
            on_shutdown: t.Callable[[CtxType], t.Awaitable[None]] = None,
            on_job_prerun: t.Optional[OnJobPrerunType] = None,
            on_job_postrun: t.Optional[OnJobPostrunType] = None,
            on_job_prepublish: t.Optional[OnJobPrepublishType] = None,
            on_scheduler_startup: t.Callable[
                [CtxType], t.Awaitable[None]] = None,
            on_scheduler_shutdown: t.Callable[
                [CtxType], t.Awaitable[None]] = None,
            max_jobs: int = 10,
            job_timeout: SecondsTimedelta = 300,
            keep_result: SecondsTimedelta = 3600,
            poll_delay: SecondsTimedelta = 0.5,
            queue_read_limit: t.Optional[int] = None,
            max_tries: int = 5,
            health_check_interval: SecondsTimedelta = 3600,
            health_check_key: t.Optional[str] = None,
            ctx: t.Optional[DataDict] = None,
            retry_jobs: bool = True,
            max_burst_jobs: int = -1,
            job_serializer: t.Optional[Serializer] = None,
            job_deserializer: t.Optional[Deserializer] = None,
            default_job_expires: AnyTimedelta = TD_1_DAY,
    ) -> None:
        self.worker_settings = WorkerSettings(
            queue_name=queue_name, burst=burst, max_jobs=max_jobs,
            job_timeout=job_timeout, keep_result=keep_result,
            poll_delay=poll_delay, queue_read_limit=queue_read_limit,
            max_tries=max_tries, health_check_interval=health_check_interval,
            health_check_key=health_check_key, retry_jobs=retry_jobs,
            max_burst_jobs=max_burst_jobs, job_serializer=job_serializer,
            job_deserializer=job_deserializer,
        )
        self.redis_pool: t.Optional['ArqRedis'] = None
        self.redis_settings = redis_settings
        self.ctx = ctx
        self.on_startup = on_startup
        self.on_shutdown = on_shutdown
        self.on_job_prerun = on_job_prerun
        self.on_job_postrun = on_job_postrun
        self.on_job_prepublish = on_job_prepublish
        self.on_scheduler_startup = on_scheduler_startup
        self.on_scheduler_shutdown = on_scheduler_shutdown
        self.default_job_expires = default_job_expires
        self.cron_jobs: t.List[CronJob] = []
        self.registry = Registry()

    async def connect(self, redis_pool: t.Optional[ArqRedis] = None) -> None:
        if self.redis_pool:
            return

        self.redis_pool = redis_pool or await create_pool(
            self.redis_settings,
            job_serializer=self.worker_settings.job_serializer,
            job_deserializer=self.worker_settings.job_deserializer,
        )

    async def disconnect(self) -> None:
        if not self.redis_pool:
            return

        self.redis_pool.close()
        await self.redis_pool.wait_closed()
        self.redis_pool = None

    def autodiscover_tasks(self, packages: t.Sequence[str]) -> None:
        for pkg in packages:
            importlib.import_module(pkg)

    def add_cron_jobs(self, *cron_jobs: CronJob) -> None:
        """
        :param cron_jobs: list of cron jobs to run,
                          use :func:`darq.cron.cron` to create them
        """
        registered_coroutines = {f.coroutine for f in self.registry.values()}
        for cj in cron_jobs:
            if not isinstance(cj, CronJob):
                raise DarqException(f'{cj!r} must be instance of CronJob')
            if cj.task not in registered_coroutines:
                raise DarqException(
                    f'{cj.task!r} is not registered. '
                    'Please, wrap it with @task decorator.',
                )
            self.cron_jobs.append(cj)

    @t.overload
    def task(
            self,
            func: WrappingFunc,
    ) -> DarqTask[WrappingFunc]:  # pragma: no cover
        ...

    @t.overload
    def task(
            self,
            *,
            keep_result: t.Optional[AnyTimedelta] = None,
            timeout: t.Optional[AnyTimedelta] = None,
            max_tries: t.Optional[int] = None,
            queue: t.Optional[str] = None,
            expires: t.Union[None, AnyTimedelta, UNSET_ARG] = unset_arg,
            with_ctx: bool = False,
    ) -> t.Callable[
        [WrappingFunc],
        DarqTask[WrappingFunc],
    ]:  # pragma: no cover
        ...

    def task(  # type: ignore[no-untyped-def]
            self,
            func=None,
            *,
            keep_result=None,
            timeout=None,
            max_tries=None,
            queue=None,
            expires=unset_arg,
            with_ctx=False,
    ):
        """
        :param func: coroutine function
        :param keep_result: duration to keep the result for, if 0 the result
                            is not kept
        :param timeout: maximum time the task should take
        :param max_tries: maximum number of tries allowed for the task,
                          use 1 to prevent retrying
        :param queue: queue of the task, can be used to send task to
                      different queue
        :param expires: if the task still hasn't started after this
                        duration, do not run it
        :param with_ctx: pass context to the task as first argument
        """
        task_queue = queue
        task_expires = expires

        def _decorate(function):  # type: ignore[no-untyped-def]
            assert asyncio.iscoroutinefunction(function), \
                f'{function} is not a coroutine function'
            name = get_function_name(function)

            async def apply_async(
                    args: t.Optional[t.Sequence[t.Any]] = None,
                    kwargs: t.Optional[t.Mapping[str, t.Any]] = None,
                    *,
                    job_id: t.Optional[str] = None,
                    queue: t.Optional[str] = None,
                    defer_until: t.Optional[datetime.datetime] = None,
                    defer_by: t.Optional[AnyTimedelta] = None,
                    expires: t.Union[None, AnyTimedelta, UNSET_ARG] = unset_arg,
                    job_try: t.Optional[int] = None,
            ) -> t.Optional[Job]:
                """
                Enqueue a task with params.

                :param args: args to pass to the function
                :param kwargs: any keyword arguments to pass to the function
                :param job_id: ID of the task, can be used to enforce task
                               uniqueness
                :param queue: queue of the task, can be used to send task to
                              different queue
                :param defer_until: datetime at which to run the task
                :param defer_by: duration to wait before running the task
                :param expires: if the task still hasn't started after this
                                duration, do not run it
                :param job_try: useful when re-enqueueing task
                :return: :class:`darq.jobs.Job` instance or ``None`` if a task
                         with this ID already exists
                """
                args = list(args) if args is not None else []
                kwargs = dict(kwargs) if kwargs is not None else {}
                queue = queue or task_queue

                if expires is unset_arg:
                    expires = (
                        task_expires
                        if task_expires is not unset_arg
                        else self.default_job_expires
                    )
                expires = t.cast(t.Optional[AnyTimedelta], expires)

                if not self.redis_pool:
                    raise DarqConnectionError(
                        'Darq app is not connected. Please, make '
                        '"await <darq_instance>.connect()" before calling '
                        'this function',
                    )
                metadata: DataDict = {}

                job_options = JobEnqueueOptions({
                    'job_id': job_id, 'queue_name': queue,
                    'defer_until': defer_until, 'defer_by': defer_by,
                    'expires': expires, 'job_try': job_try,
                })

                if self.on_job_prepublish:
                    await self.on_job_prepublish(
                        metadata, self.registry[name], args, kwargs,
                        job_options,
                    )
                if metadata:
                    kwargs['__metadata__'] = metadata
                return await self.redis_pool.enqueue_job(
                    name, tuple(args), kwargs, **job_options,
                )

            async def delay(*args: t.Any, **kwargs: t.Any) -> t.Optional[Job]:
                return await apply_async(args, kwargs)

            function.delay = delay
            function.apply_async = apply_async
            self.registry.add(Task.new(
                coroutine=function, name=name,
                keep_result=keep_result, timeout=timeout, max_tries=max_tries,
                with_ctx=with_ctx,
            ))

            return function

        if func:
            return _decorate(func)  # type: ignore[no-untyped-call]

        return _decorate
